"""Parse a regular expression defined over a possibly-formatted string."""
import collections.abc
import io
import math
import re
from typing import Tuple, Optional, List, Sequence, Union, MutableMapping

from icontract import invariant, require, ensure, snapshot

from aas_core_codegen.common import pairwise
from aas_core_codegen.parse.retree._types import (
    Char,
    Range,
    Concatenation,
    Symbol,
    Group,
    CharSet,
    Quantifier,
    Term,
    UnionExpr,
    Regex,
    TermValueUnion,
    SymbolKind,
)
from aas_core_codegen.parse.tree import FormattedValue


# fmt: off
@invariant(
    lambda self:
    (
        not self.done()
        ^ (
            self.major_cursor == len(self.values)
            or (
                self.major_cursor == len(self.values) - 1
                and isinstance(self.pointed_value(), str)
                and self.minor_cursor == len(self.pointed_value())
            )
        )
    ),
    "We are done iff we reached the end of values with the major cursor or "
    "the end of the last pointed value as string"
)
@invariant(
    lambda self:
    not isinstance(self.pointed_value(), str)
    or (0 <= self.minor_cursor <= len(self.pointed_value())),
    "Minor cursor is within the expected range if the pointed value is a string"
)
@invariant(
    lambda self:
    not (not isinstance(self.pointed_value(), str))
    or (self.minor_cursor is None),
    "Minor cursor is invalidated if the pointed value is not a string"
)
@invariant(
    lambda self:
    (
            not (self.major_cursor == len(self.values))
            or self.pointed_value() is None
    ),
    "No pointed value if we reached the end of the values"
)
@invariant(
    lambda self:
    (
            not (0 <= self.major_cursor < len(self.values))
            or self.pointed_value() is not None
    ),
    "A value must be pointed at if we did not reach the end of the values"
)
@invariant(
    lambda self:
    0 <= self.major_cursor <= len(self.values),
    "Major cursor is within the expected range"
)
# fmt: on
class Cursor:
    """Manage the iteration over the input stream."""

    # fmt: off
    @require(
        lambda values:
        not (len(values) > 1)
        or not any(
            (
                    isinstance(first, str)
                    and isinstance(second, str)
            )
            for first, second in zip(values, values[1:])
        ),
        "No consecutive strings expected in the parsed sequence of an AST expression "
        "representing a string interpolation"
    )
    # NOTE (mristin, 2024-05-08):
    # We add a runtime test here since it already happened that we supplied ``values``
    # as a string during development, causing an unnecessary long debugging session.
    @require(
        lambda values:
        isinstance(values, collections.abc.Sequence),
        "values must be a sequence"
    )
    # fmt: on
    def __init__(self, values: Sequence[Union[str, FormattedValue]]) -> None:
        """Initialize with the given values."""
        self.values = values

        # Major cursor points within the values
        self._major_cursor = 0

        # Minor cursor points within a value
        self._minor_cursor: Optional[int] = None
        if isinstance(self.pointed_value(), str):
            self._minor_cursor = 0

    def copy(self) -> "Cursor":
        """Make a deep copy of the cursor."""
        return Cursor(values=self.values)

    @property
    def major_cursor(self) -> int:
        """Return the current cursor in the :attr:`~values`."""
        return self._major_cursor

    @property
    def minor_cursor(self) -> Optional[int]:
        """
        Return the current cursor in the :attr:`~pointed_value`.

        If the :attr:`~pointed_value` is not a string, return ``None``.
        """
        if not isinstance(self.pointed_value(), str):
            return None

        return self._minor_cursor

    def pointed_value(self) -> Optional[Union[str, FormattedValue]]:
        """Return the value pointed by the :attr:`~major_cursor`."""
        if self._major_cursor >= len(self.values):
            return None

        return self.values[self._major_cursor]

    def _move_major_cursor(self) -> None:
        """
        Move the major cursor to the next value.

        The minor cursor is also updated accordingly, depending on the pointed value.
        """
        if not self.done():
            self._major_cursor += 1

            self._minor_cursor = 0 if isinstance(self.pointed_value(), str) else None

    @require(lambda jump_length: jump_length >= 0)
    @require(
        lambda self: not (not self.done())
        or (self._minor_cursor is not None and isinstance(self.pointed_value(), str)),
        "If we are not done, we should only move within a string value",
    )
    def _move_minor_cursor(self, jump_length: int) -> None:
        """
        Move the minor cursor within a string value for ``jump_length`` positions.

        If we reached the end of value, move to the next value.
        """
        if self.done():
            return

        assert self._minor_cursor is not None

        pointed_value = self.pointed_value()
        assert pointed_value is not None
        assert isinstance(pointed_value, str)

        if self._minor_cursor + jump_length >= len(pointed_value):
            self._move_major_cursor()
        else:
            self._minor_cursor += jump_length

    def done(self) -> bool:
        """Signal that we reached the end of the :attr:`~values`."""
        if self._major_cursor >= len(self.values):
            return True

        pointed_value = self.pointed_value()
        if isinstance(pointed_value, str):
            assert self._minor_cursor is not None
            return self._minor_cursor >= len(pointed_value)
        else:
            assert self._minor_cursor is None
            return False

    def try_formatted_value(self) -> Optional[FormattedValue]:
        """
        Try to read the formatted value at the current cursor and return it.

        Otherwise, if no formatted value is present, return ``None``.
        """
        value = self.pointed_value()
        if isinstance(value, FormattedValue):
            self._move_major_cursor()
            return value

        return None

    def try_positive_integer_without_sign(self) -> Optional[int]:
        """
        Try to read an integer and move the cursor accordingly.

        An integer can be prefixed by a zero (*e.g.*, ``000123``). No sign is expected
        as a prefix.

        Otherwise, if there is no integer at the cursor, return ``None`` and do not
        move the cursor.
        """
        pointed_value = self.pointed_value()
        if isinstance(pointed_value, FormattedValue):
            return None

        if self.minor_cursor is None:
            return None

        assert isinstance(pointed_value, str)
        accumulator = []  # type: List[str]
        for i in range(self.minor_cursor, len(pointed_value)):
            if pointed_value[i].isdigit():
                accumulator.append(pointed_value[i])
            else:
                break

        if len(accumulator) > 0:
            self._move_minor_cursor(jump_length=len(accumulator))
            return int("".join(accumulator))

        return None

    @snapshot(lambda self, literal: self.peek_literal(literal), "peeked")
    @ensure(
        lambda result, OLD: (result and OLD.peeked) or (not result and not OLD.peeked),
        "If we matched the literal, we should have been able to successfully peek it "
        "before moving the cursor, and vice-versa.",
    )
    def try_literal(self, literal: str) -> bool:
        """
        Try to match the ``literal``, return ``True`` and move the cursor accordingly.

        If the ``literal`` could not be matched, return ``False`` and do not move the
        cursor.
        """
        result = self.peek_literal(literal=literal)

        if result:
            self._move_minor_cursor(jump_length=len(literal))

        return result

    def peek_literal(self, literal: str) -> bool:
        """
        Try to match the ``literal``.

        Do not move the cursor.
        """
        pointed_value = self.pointed_value()
        if not isinstance(pointed_value, str):
            return False

        assert self.minor_cursor is not None
        if len(pointed_value) < self.minor_cursor + len(literal):
            return False

        result = (
            pointed_value[self.minor_cursor : (self.minor_cursor + len(literal))]
            == literal
        )

        return result

    def peek_substring(self, length: int) -> Optional[str]:
        """
        Try to read a substring of ``length`` characters.

        If the remainder of the stream contains less than ``length`` characters,
        return ``None``, or the pointed value is not a string, return ``None``.

        Do not move the cursor.
        """
        pointed_value = self.pointed_value()
        if not isinstance(pointed_value, str):
            return None

        assert self.minor_cursor is not None
        if len(pointed_value) < self.minor_cursor + length:
            return None

        return pointed_value[self.minor_cursor : (self.minor_cursor + length)]

    def try_substring(self, length: int) -> Optional[str]:
        """
        Try to read a substring of ``length`` characters and move the cursor.

        If the remainder of the stream contains less than ``length`` characters,
        or the pointed value is not a string, return ``None`` and
        do not move the cursor.
        """
        result = self.peek_substring(length=length)
        if result is None:
            return None

        self._move_minor_cursor(jump_length=length)

        return result

    def try_spaces_or_tabs(self) -> bool:
        """
        Try to read spaces (``" "``) and tabs (``"\t"``) from the stream.

        If the spaces are found, move the cursor and return ``True``.

        If no spaces are found at the cursor, return ``False`` and do not move
        the cursor.
        """
        found_tab_or_space = False
        while True:
            if self.try_literal(" "):
                found_tab_or_space = True
            elif self.try_literal("\t"):
                found_tab_or_space = True
            else:
                break

        return found_tab_or_space

    # fmt: off
    @require(
        lambda other, self:
        id(other.values) == id(self.values),
        "This and the other cursor must point to the same values"
    )
    # fmt: on
    def has_same_position(self, other: "Cursor") -> bool:
        """Check whether the ``other`` cursor points to the same position."""
        return (self.done() and other.done()) or (
            self.major_cursor == other.major_cursor
            and self.minor_cursor == other.minor_cursor
        )

    # fmt: off
    @require(
        lambda other, self:
        id(other.values) == id(self.values),
        "This and the other cursor must point to the same values"
    )
    @ensure(
        lambda other, self, result:
        not (self.has_same_position(other))
        or (not result)
    )
    # fmt: on
    def is_before(self, other: "Cursor") -> bool:
        """Check whether this cursor is pointing before the ``other`` cursor."""
        if self.done() and other.done():
            return False
        elif self.done() and not other.done():
            return False
        elif not self.done() and other.done():
            return True
        else:
            assert self.major_cursor is not None
            assert other.major_cursor is not None

            if self.major_cursor < other.major_cursor:
                return True
            elif self.major_cursor > other.major_cursor:
                return False
            else:
                if self.minor_cursor is None and other.minor_cursor is None:
                    # NOTE (mristin, 2022-06-02):
                    # Both major cursors point to the formatted value, so they
                    # point to the same position.
                    assert isinstance(self.pointed_value(), FormattedValue)
                    assert isinstance(other.pointed_value(), FormattedValue)
                    return False
                else:
                    # NOTE (mristin, 2022-06-02):
                    # If one minor cursor is not None, it means the major cursors
                    # point to a string.
                    assert isinstance(self.pointed_value(), str)
                    assert isinstance(other.pointed_value(), str)

                    assert self.minor_cursor is not None
                    assert other.minor_cursor is not None
                    return self.minor_cursor < other.minor_cursor


class Error:
    """Represent an error occurring during the parsing of a regular expression."""

    def __init__(self, message: str, cursor: Cursor) -> None:
        """Initialize with the given values."""
        self.message = message
        self.cursor = cursor


_UNEXPECTED_CHARACTERS_IN_REGEX_RE = re.compile("[\\n\\f\\v\\r]")
_WHITESPACE_RE = re.compile("\\s")


@require(lambda cursor: len(cursor.values) > 0)
@ensure(lambda result: len(result[0]) > 0 and len(result[1]) > 0)
@ensure(lambda result: result[1].endswith("^"))
def render_pointer(cursor: Cursor) -> Tuple[str, str]:
    """Render the regular expression and the pointer corresponding to the cursor."""
    regex_writer = io.StringIO()
    pointer_writer = io.StringIO()

    @require(lambda text: not _UNEXPECTED_CHARACTERS_IN_REGEX_RE.search(text))
    def write_text(text: str, draw_pointer: bool) -> None:
        """
        Output the ``text``.

        We write the ``text`` to the ``regex_writer``. Depending on ``draw_pointer``,
        we either write the corresponding whitespace to ``pointer_writer``
        (if ``False``) or draw a pointer (``^``).
        """
        if not draw_pointer:
            for character in text:
                regex_writer.write(character)

                if _WHITESPACE_RE.fullmatch(character):
                    pointer_writer.write(character)
                else:
                    pointer_writer.write(" ")
        else:
            regex_writer.write(text)
            pointer_writer.write("^")

    a_cursor = Cursor(values=cursor.values)
    while True:
        if a_cursor.done():
            assert a_cursor.has_same_position(cursor)
            pointer_writer.write("^")
            break

        has_same_position = a_cursor.has_same_position(cursor)

        formatted_value = a_cursor.try_formatted_value()
        if formatted_value:
            # NOTE (mristin, 2022-06-09):
            # The asttokens can not get the text of the formatted values,
            # see: https://github.com/gristlabs/asttokens/issues/6.
            write_text(text="<formatted value>", draw_pointer=has_same_position)
        else:
            substring = a_cursor.try_substring(length=1)
            assert substring is not None, (
                f"We explicitly checked for a_cursor "
                f"not done and not a formatted value, but got no substring: "
                f"{a_cursor.pointed_value()=}, {a_cursor.major_cursor=}, "
                f"{a_cursor.minor_cursor=}"
            )
            write_text(text=substring, draw_pointer=has_same_position)

        if has_same_position:
            break

    # NOTE (mristin, 2022-06-09):
    # Write the remainder of the regular expression
    while not a_cursor.done():
        formatted_value = a_cursor.try_formatted_value()
        if formatted_value:
            # NOTE (mristin, 2022-06-09):
            # The asttokens can not get the text of the formatted values,
            # see: https://github.com/gristlabs/asttokens/issues/6.
            regex_writer.write("<formatted value>")
        else:
            substring = a_cursor.try_substring(length=1)
            assert substring is not None, (
                f"We explicitly checked for a_cursor "
                f"not done and not a formatted value, but got no substring: "
                f"{a_cursor.pointed_value()=}, {a_cursor.major_cursor=}, "
                f"{a_cursor.minor_cursor=}"
            )
            regex_writer.write(substring)

    return regex_writer.getvalue(), pointer_writer.getvalue()


# fmt: off
# noinspection RegExpSimplifiable
@require(
    lambda cursor: not cursor.peek_literal("-"),
    "The dash (``-``) handled before, outside of this function"
)
@require(
    lambda cursor: not cursor.done(),
    "End of input handled before, outside of this function"
)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
# fmt: on
def _parse_range_char(cursor: Cursor) -> Tuple[Optional[Char], Optional[Error]]:
    """Try to parse a character in the range of a character set."""
    if cursor.try_literal("\\x"):
        substring = cursor.try_substring(length=2)
        if substring is None:
            return None, Error("Expected two hexadecimal digits after \\x", cursor)

        # noinspection RegExpSimplifiable
        if not re.fullmatch(r"[a-fA-F0-9]{2}", substring):
            return None, Error(
                f"Expected two hexadecimal digits after \\x, but got {substring!r}",
                cursor,
            )

        result = Char(character=chr(int(substring, 16)), explicitly_encoded=True)

    elif cursor.try_literal("\\u"):
        substring = cursor.try_substring(length=4)
        if substring is None:
            return None, Error("Expected four hexadecimal digits after \\u", cursor)

        if not re.fullmatch(r"[a-fA-F0-9]{4}", substring):
            return None, Error(
                f"Expected four hexadecimal digits after \\u, "
                f"but got {substring!r}",
                cursor,
            )

        result = Char(character=chr(int(substring, 16)), explicitly_encoded=True)

    elif cursor.try_literal("\\U"):
        substring = cursor.try_substring(length=8)
        if substring is None:
            return None, Error("Expected eight hexadecimal digits after \\U", cursor)

        if not re.fullmatch(r"[a-fA-F0-9]{8}", substring):
            return None, Error(
                f"Expected eight hexadecimal digits after \\U, "
                f"but got {substring!r}",
                cursor,
            )

        code = int(substring, 16)
        if code < 0x00010000 or code > 0x0010FFFF:
            # noinspection SpellCheckingInspection
            return None, Error(
                f"Expected the character code in the range \\U00010000-\\U0010FFFF, "
                f"but got \\U{code:08x}; "
                f"the code can not be represented with high/low surrogates in UTF-16",
                cursor,
            )

        result = Char(character=chr(code), explicitly_encoded=True)
    elif cursor.try_literal("\\t"):
        result = Char(character="\t")

    elif cursor.try_literal("\\n"):
        result = Char(character="\n")

    elif cursor.try_literal("\\r"):
        result = Char(character="\r")

    elif cursor.try_literal("\\f"):
        result = Char(character="\f")

    elif cursor.try_literal("\\v"):
        result = Char(character="\v")

    elif cursor.try_literal("\\\\"):
        result = Char(character="\\")

    elif cursor.try_literal("\\["):
        result = Char(character="[")

    elif cursor.try_literal("\\]"):
        result = Char(character="]")

    elif cursor.try_literal("\\^"):
        result = Char(character="^")

    elif cursor.try_literal("\\-"):
        result = Char(character="-")

    elif cursor.try_literal("\\s") or cursor.try_literal("\\S"):
        return None, Error(
            "We do not handle whitespace escaping at the moment (2022-06-09), "
            "as the notion of a whitespace depends on the regex engine",
            cursor,
        )

    elif cursor.try_literal("\\d") or cursor.try_literal("\\D"):
        return None, Error(
            "We do not handle digit escaping at the moment (2022-06-09), "
            "as the notion of a digit depends on the regex engine",
            cursor,
        )

    elif cursor.try_literal("\\"):
        return None, Error("Unexpected escaping", cursor)

    else:
        character = cursor.try_substring(length=1)
        assert character is not None, (
            "Expected to read at least one character since we explicitly checked "
            "that the cursor is not done before"
        )
        result = Char(character=character)

    assert result is not None
    return result, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _parse_ranges_and_closing(
    cursor: Cursor,
) -> Tuple[Optional[List[Range]], Optional[Error]]:
    # noinspection GrazieInspection
    """Parse one or more ranges in a character set and the closing ``]``."""
    ranges = []  # type: List[Range]

    # Prefix dash is allowed and should be considered as a character.
    if cursor.try_literal("-"):
        ranges.append(Range(start=Char("-"), end=None))

    cursor_by_range = dict()  # type: MutableMapping[Range, Cursor]

    while True:
        if cursor.done():
            return None, Error(
                "Expected a closing bracket for the character set (``]``), "
                "but reached the end of input.",
                cursor,
            )

        cursor_at_start = cursor.copy()

        # NOTE (mristin, 2022-06-08):
        # A suffix dash is also allowed and should be considered a single character.
        if cursor.try_literal("-]"):
            the_range = Range(start=Char("-"), end=None)
            cursor_by_range[the_range] = cursor_at_start
            ranges.append(the_range)
            break
        elif cursor.try_literal("]"):
            break
        else:
            start, error = _parse_range_char(cursor)
            if error is not None:
                return None, error
            assert start is not None

            end = None  # type: Optional[Char]
            if not cursor.peek_literal("-]"):
                if cursor.try_literal("-"):
                    if cursor.try_literal("-"):
                        return None, Error(
                            "Unexpected double ``-`` in the character set", cursor
                        )

                    end, error = _parse_range_char(cursor)
                    if error is not None:
                        return None, error
                    assert end is not None

            if end is not None and ord(start.character) > ord(end.character):
                return None, Error(
                    "Invalid character range, start is smaller than end", cursor
                )

            the_range = Range(start=start, end=end)
            cursor_by_range[the_range] = cursor_at_start
            ranges.append(the_range)

    for this_range, next_range in pairwise(
        sorted(ranges, key=lambda rng: ord(rng.start.character))
    ):
        this_end = this_range.end if this_range.end is not None else this_range.start

        next_start = next_range.start

        if ord(this_end.character) >= ord(next_start.character):
            this_range_str = (
                f"{this_range.start.character!r}"
                if this_range.end is None
                else f"{this_range.start.character!r}-{this_range.end.character!r}"
            )

            next_range_str = (
                f"{next_range.start.character!r}"
                if next_range.end is None
                else f"{next_range.start.character!r}-{next_range.end.character!r}"
            )

            return None, Error(
                f"The range {this_range_str} and the range {next_range_str} overlap",
                cursor_by_range[this_range],
            )

    return ranges, None


# noinspection RegExpSimplifiable
def _parse_char_literal(cursor: Cursor) -> Tuple[Optional[Char], Optional[Error]]:
    """
    Parse a character literal in a concatenation.

    It might be that no character could be parsed, but that the stream is still valid.
    For example, if there is a quantifier (such as ``*``, ``+`` or ``?``) following
    a literal. In those cases, return ``None`` and no error.
    """
    if cursor.done():
        return None, None

    result: Char

    if cursor.try_literal("\\x"):
        substring = cursor.try_substring(length=2)
        if substring is None:
            return None, Error("Expected two hexadecimal digits after \\x", cursor)

        if not re.fullmatch(r"[a-fA-F0-9]{2}", substring):
            return None, Error(
                f"Expected two hexadecimal digits after \\x, but got {substring!r}",
                cursor,
            )

        result = Char(character=chr(int(substring, 16)), explicitly_encoded=True)

    elif cursor.try_literal("\\u"):
        substring = cursor.try_substring(length=4)
        if substring is None:
            return None, Error("Expected four hexadecimal digits after \\u", cursor)

        if not re.fullmatch(r"[a-fA-F0-9]{4}", substring):
            return None, Error(
                f"Expected four hexadecimal digits after \\u, "
                f"but got {substring!r}",
                cursor,
            )

        result = Char(character=chr(int(substring, 16)), explicitly_encoded=True)

    elif cursor.try_literal("\\U"):
        substring = cursor.try_substring(length=8)
        if substring is None:
            return None, Error("Expected eight hexadecimal digits after \\U", cursor)

        if not re.fullmatch(r"[a-fA-F0-9]{8}", substring):
            return None, Error(
                f"Expected eight hexadecimal digits after \\U, "
                f"but got {substring!r}",
                cursor,
            )

        code = int(substring, 16)
        if code < 0x00010000 or code > 0x0010FFFF:
            # noinspection SpellCheckingInspection
            return None, Error(
                f"Expected the character code in the range \\U00010000-\\U0010FFFF, "
                f"but got \\U{code:08x}; "
                f"the code can not be represented with high/low surrogates in UTF-16",
                cursor,
            )

        result = Char(character=chr(code), explicitly_encoded=True)
    elif cursor.try_literal("\\t"):
        result = Char(character="\t")

    elif cursor.try_literal("\\n"):
        result = Char(character="\n")

    elif cursor.try_literal("\\r"):
        result = Char(character="\r")

    elif cursor.try_literal("\\f"):
        result = Char(character="\f")

    elif cursor.try_literal("\\v"):
        result = Char(character="\v")

    elif cursor.try_literal("\\."):
        result = Char(character=".")

    elif cursor.try_literal("\\#"):
        result = Char(character="#")

    elif cursor.try_literal("\\^"):
        result = Char(character="^")

    elif cursor.try_literal("\\$"):
        result = Char(character="$")

    elif cursor.try_literal("\\("):
        result = Char(character="(")

    elif cursor.try_literal("\\)"):
        result = Char(character=")")

    elif cursor.try_literal("\\["):
        result = Char(character="[")

    elif cursor.try_literal("\\]"):
        result = Char(character="]")

    elif cursor.try_literal("\\\\"):
        result = Char(character="\\")

    elif cursor.try_literal("\\*"):
        result = Char(character="*")

    elif cursor.try_literal("\\+"):
        result = Char(character="+")

    elif cursor.try_literal("\\?"):
        result = Char(character="?")

    elif cursor.try_literal("\\s") or cursor.try_literal("\\S"):
        return None, Error(
            "We do not handle whitespace escaping at the moment (2022-06-09), "
            "as the notion of a whitespace depends on the regex engine",
            cursor,
        )

    elif cursor.try_literal("\\w") or cursor.try_literal("\\W"):
        return None, Error(
            "We do not handle word escaping at the moment (2022-06-09), "
            "as the notion of a word depends on the regex engine",
            cursor,
        )

    elif cursor.try_literal("\\d") or cursor.try_literal("\\D"):
        return None, Error(
            "We do not handle digit escaping at the moment (2022-06-09), "
            "as the notion of a digit depends on the regex engine",
            cursor,
        )

    elif cursor.try_literal("\\"):
        return None, Error("Unexpected escaping", cursor)

    elif cursor.try_literal("^"):
        raise AssertionError(
            "Expected to handle the start symbol before, outside of this function"
        )

    elif cursor.try_literal("$"):
        raise AssertionError(
            "Expected to handle the end symbol before, outside of this function"
        )

    elif cursor.try_literal("*"):
        raise AssertionError(
            "Expected to handle the ``*`` quantifier before, outside of this function"
        )

    elif cursor.try_literal("+"):
        raise AssertionError(
            "Expected to handle the ``+`` quantifier before, outside of this function"
        )

    elif cursor.try_literal("?"):
        raise AssertionError(
            "Expected to handle the ``?`` quantifier before, outside of this function"
        )

    elif cursor.try_literal("{"):
        raise AssertionError(
            "Expected to handle the opening ``{`` of a quantifier before, "
            "outside of this function"
        )

    elif cursor.peek_literal(")") or cursor.peek_literal("|"):
        # NOTE (mristin, 2022-06-08):
        # We encountered a closing bracket or a delimiter in a union,
        # so no concatenation is possible anymore, and we need to match an "empty"
        # character literal.
        return None, None

    else:
        character = cursor.try_substring(length=1)
        assert character is not None, (
            "Expected to read at least one character since we explicitly checked "
            "that the cursor is not done before"
        )
        result = Char(character=character)

    return result, None


_SUPPLEMENTARY_PLANE_START = 0x00010000


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _parse_concatenation(
    cursor: Cursor,
) -> Tuple[Optional[Concatenation], Optional[Error]]:
    """Try to parse a concatenation of regex terms."""
    if cursor.done():
        return None, Error(
            "Expected at least one term, but reached the end of input.", cursor
        )

    concatenants = []  # type: List[Term]

    while True:
        old_cursor_position = (
            cursor.major_cursor,
            cursor.minor_cursor if cursor.minor_cursor is not None else math.inf,
        )

        if cursor.done():
            break

        if cursor.peek_literal("|"):
            break

        value: Optional[TermValueUnion]

        # noinspection GrazieInspection
        if cursor.try_literal("^"):
            value = Symbol(kind=SymbolKind.START)

        elif cursor.try_literal("$"):
            value = Symbol(kind=SymbolKind.END)

        elif cursor.try_literal("."):
            value = Symbol(kind=SymbolKind.DOT)

        elif cursor.try_literal("("):
            if cursor.try_literal("?"):
                return None, Error(
                    "At this moment (2022-06-09), we did not have time to implement "
                    "the support for the directives in the groups (``?``). "
                    "Please inform the developers if you need this feature.",
                    cursor,
                )

            union, error = _parse_union(cursor)
            if error is not None:
                return None, error
            assert union is not None

            found_closing_bracket = cursor.try_literal(")")
            if not found_closing_bracket:
                return None, Error(
                    "Expected a closing bracket for the group, but found none.", cursor
                )

            value = Group(union=union)
        elif cursor.try_literal("[^"):
            ranges, error = _parse_ranges_and_closing(cursor)
            if error is not None:
                return None, error
            assert ranges is not None

            for a_range in ranges:
                if ord(a_range.start.character) > _SUPPLEMENTARY_PLANE_START or (
                    a_range.end is not None
                    and ord(a_range.end.character) > _SUPPLEMENTARY_PLANE_START
                ):
                    range_str = (
                        f"\\U{a_range.start:08x}"
                        if a_range.end is None
                        else f"\\U{a_range.start:08x}-\\U{a_range.end:08x}"
                    )
                    return None, Error(
                        f"Complementing character sets with a range {range_str} "
                        f"involving UTF-32 characters can not be supported, "
                        f"since we can not represent them in implementations relying "
                        f"on UTF-16-only regex engines (such as C# as of 2022-06-11)",
                        cursor,
                    )

            value = CharSet(complementing=True, ranges=ranges)

        elif cursor.try_literal("["):
            ranges, error = _parse_ranges_and_closing(cursor)
            if error is not None:
                return None, error
            assert ranges is not None

            value = CharSet(complementing=False, ranges=ranges)

        elif isinstance(cursor.pointed_value(), FormattedValue):
            formatted_value = cursor.try_formatted_value()
            assert formatted_value is not None

            value = formatted_value

        elif cursor.try_literal("*"):
            return None, Error("Unexpected ``*`` quantifier without a term", cursor)

        elif cursor.try_literal("+"):
            return None, Error("Unexpected ``+`` quantifier without a term", cursor)

        elif cursor.try_literal("?"):
            return None, Error("Unexpected ``?`` quantifier without a term", cursor)

        else:
            value, error = _parse_char_literal(cursor)
            if error is not None:
                return None, error

            # NOTE (mristin, 2022-06-08):
            # The ``value`` can be None here. For example, if we peeked a closing
            # ``)`` or a delimiting ``|``.

        if value is not None:
            quantifier = None  # type: Optional[Quantifier]
            if cursor.try_literal("*?"):
                quantifier = Quantifier(non_greedy=True, minimum=0, maximum=None)
            elif cursor.try_literal("+?"):
                quantifier = Quantifier(non_greedy=True, minimum=1, maximum=None)
            elif cursor.try_literal("??"):
                quantifier = Quantifier(non_greedy=True, minimum=0, maximum=1)
            elif cursor.try_literal("*"):
                quantifier = Quantifier(non_greedy=False, minimum=0, maximum=None)
            elif cursor.try_literal("+"):
                quantifier = Quantifier(non_greedy=False, minimum=1, maximum=None)
            elif cursor.try_literal("?"):
                quantifier = Quantifier(non_greedy=False, minimum=0, maximum=1)
            elif cursor.try_literal("{"):
                cursor.try_spaces_or_tabs()
                minimum = cursor.try_positive_integer_without_sign()
                cursor.try_spaces_or_tabs()
                found_comma = cursor.try_literal(",")
                cursor.try_spaces_or_tabs()
                maximum = cursor.try_positive_integer_without_sign()
                cursor.try_spaces_or_tabs()

                if minimum is None and maximum is None:
                    return None, Error(
                        "Encountered an opening ``{`` for a quantifier, "
                        "but both minimum and the maximum could not be parsed",
                        cursor,
                    )

                if not found_comma:
                    maximum = minimum

                minimum_zero_if_none = minimum if minimum is not None else 0

                if cursor.try_literal("}?"):
                    quantifier = Quantifier(
                        non_greedy=True, minimum=minimum_zero_if_none, maximum=maximum
                    )
                elif cursor.try_literal("}"):
                    quantifier = Quantifier(
                        non_greedy=False, minimum=minimum_zero_if_none, maximum=maximum
                    )
                else:
                    return None, Error(
                        "Expected closing ``}`` for the quantifier", cursor
                    )
            else:
                pass

            if isinstance(value, Symbol) and (
                value.kind is SymbolKind.START or value.kind is SymbolKind.END
            ):
                assert quantifier is None, (
                    f"Unexpected the symbol {value.kind.value} with a quantifier. "
                    f"This should have been detected before."
                )

            concatenants.append(Term(value=value, quantifier=quantifier))
        else:
            break

        cursor_position = (
            cursor.major_cursor,
            cursor.minor_cursor if cursor.minor_cursor is not None else math.inf,
        )
        assert (
            old_cursor_position < cursor_position
        ), f"Loop invariant: {old_cursor_position=}, {cursor_position=}"

    # NOTE (mristin, 2022-06-03):
    # Empty terms are possible! For example, parsing ``(())`` needs to allow for an
    # empty union expression in the inner group.

    return Concatenation(concatenants=concatenants), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _parse_union(cursor: Cursor) -> Tuple[Optional[UnionExpr], Optional[Error]]:
    """Try to parse a union of concatenations."""
    if cursor.done():
        return UnionExpr(uniates=[]), None

    concatenation, error = _parse_concatenation(cursor)
    if error is not None:
        return None, error
    assert concatenation is not None

    uniates = [concatenation]

    while cursor.try_literal(literal="|"):
        if cursor.done():
            # NOTE (mristin, 2022-06-09):
            # An empty term at the end of input is allowed.
            # For example, consider ``"|"``.
            uniates.append(Concatenation(concatenants=[]))
            break

        concatenation, error = _parse_concatenation(cursor)
        if error is not None:
            return None, error
        assert concatenation is not None

        uniates.append(concatenation)

    return UnionExpr(uniates=uniates), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _parse_regex(cursor: Cursor) -> Tuple[Optional[Regex], Optional[Error]]:
    """Try to parse the regular expression and move the cursor."""
    union, error = _parse_union(cursor)
    if error is not None:
        return None, error

    if not cursor.done():
        return None, Error(
            message=(
                "Expected to reach the end of input, "
                "but there is still some unconsumed input left"
            ),
            cursor=cursor,
        )

    assert union is not None
    return Regex(union=union), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def parse(
    values: Sequence[Union[str, FormattedValue]]
) -> Tuple[Optional[Regex], Optional[Error]]:
    """Try to parse the ``values`` into an AST representing the regular expression."""
    cursor = Cursor(values=values)
    parsed, error = _parse_regex(cursor)
    return parsed, error



"""Transpile meta-model Python code to C++ code."""
import abc
import io
from typing import (
    Tuple,
    Optional,
    List,
    Mapping,
    Union,
    Set,
)

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Error,
    Stripped,
    assert_never,
    indent_but_first_line,
    Identifier,
)
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
)
from aas_core_codegen.intermediate import type_inference as intermediate_type_inference
from aas_core_codegen.parse import tree as parse_tree
from aas_core_codegen.cpp import (
    common as cpp_common,
    naming as cpp_naming,
)


# fmt: off
@require(
    lambda type_annotation:
    not isinstance(type_annotation, intermediate_type_inference.OurTypeAnnotation)
    or isinstance(type_annotation.our_type, intermediate.ConstrainedPrimitive)
)
# fmt: on
def _determine_which_to_wstring(
    type_annotation: Union[
        intermediate_type_inference.PrimitiveTypeAnnotation,
        intermediate_type_inference.OurTypeAnnotation,
    ]
) -> Optional[str]:
    """
    Determine which to-wstring function should be used.

    None indicates that the value needs not be converted (*i.e.*, it is already
    a wstring).
    """
    if isinstance(type_annotation, intermediate_type_inference.PrimitiveTypeAnnotation):
        if type_annotation.a_type is intermediate_type_inference.PrimitiveType.STR:
            return None
        elif type_annotation.a_type is intermediate_type_inference.PrimitiveType.INT:
            return "std::to_wstring"
        elif type_annotation.a_type is intermediate_type_inference.PrimitiveType.FLOAT:
            return "std::to_wstring"
        elif (
            type_annotation.a_type
            is intermediate_type_inference.PrimitiveType.BYTEARRAY
        ):
            base64_encode = cpp_naming.function_name(Identifier("base64_encode"))
            return f"wstringification::{base64_encode}"
        else:
            return "wstringification::to_wstring"

    elif isinstance(
        type_annotation, intermediate_type_inference.OurTypeAnnotation
    ) and isinstance(type_annotation.our_type, intermediate.ConstrainedPrimitive):
        constrainee = type_annotation.our_type.constrainee

        if constrainee is intermediate.PrimitiveType.BOOL:
            return "wstringification::to_wstring"
        elif constrainee is intermediate.PrimitiveType.INT:
            return "std::to_wstring"
        elif constrainee is intermediate.PrimitiveType.FLOAT:
            return "std::to_wstring"
        elif constrainee is intermediate.PrimitiveType.STR:
            return None
        elif constrainee is intermediate.PrimitiveType.BYTEARRAY:
            base64_encode = cpp_naming.function_name(Identifier("base64_encode"))
            return f"wstringification::{base64_encode}"
        else:
            assert_never(constrainee)

    else:
        raise ValueError(
            f"Unexpected type annotation for which we can not determine "
            f"the to-wstring function: {type_annotation!r}"
        )


# NOTE (mristin, 2023-07-01):
# We have to implement a very similar function for generating type annotations to
# ``aas_core_codegen.cpp.common.generate_type`` since we can not simply pass
# ``intermediate_type_inference.TypeAnnotationUnion`` to
# ``aas_core_codegen.cpp.common.generate_type``.

PRIMITIVE_TYPE_MAP = {
    intermediate_type_inference.PrimitiveType.BOOL: Stripped("bool"),
    intermediate_type_inference.PrimitiveType.INT: Stripped("int64_t"),
    intermediate_type_inference.PrimitiveType.FLOAT: Stripped("double"),
    intermediate_type_inference.PrimitiveType.STR: Stripped("std::wstring"),
    intermediate_type_inference.PrimitiveType.BYTEARRAY: Stripped(
        "std::vector<std::uint8_t>"
    ),
    intermediate_type_inference.PrimitiveType.NONE: Stripped("void*"),
    intermediate_type_inference.PrimitiveType.LENGTH: Stripped("size_t"),
}


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def generate_type(
    type_annotation: intermediate_type_inference.TypeAnnotationUnion,
    types_namespace: Optional[Identifier] = None,
) -> Tuple[Optional[Stripped], Optional[str]]:
    """
    Generate the C++ type for the given type annotation.

    If ``types_namespace`` is specified, it is prepended to all our types.

    (mristin, 2023-07-01): We do not handle all the type annotations from
    :py:mod:`aas_core_codegen.intermediate.type_inference` as that would be
    YAGNI (*e.g.*, verification functions, built-in functions *etc.*).
    If we do not know how to generate the type in C++, we return an error message.
    """
    if isinstance(type_annotation, intermediate_type_inference.PrimitiveTypeAnnotation):
        return PRIMITIVE_TYPE_MAP[type_annotation.a_type], None

    elif isinstance(type_annotation, intermediate_type_inference.OurTypeAnnotation):
        our_type = type_annotation.our_type

        if isinstance(our_type, intermediate.Enumeration):
            enum_name = cpp_naming.enum_name(type_annotation.our_type.name)
            if types_namespace is None:
                return enum_name, None

            return Stripped(f"{types_namespace}::{enum_name}"), None

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            return cpp_common.PRIMITIVE_TYPE_MAP[our_type.constrainee], None

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            # NOTE (mristin, 2023-07-01):
            # We always refer to interfaces even in cases of concrete classes without
            # concrete descendants since we want to allow enhancing.
            interface_name = cpp_naming.interface_name(our_type.name)

            if types_namespace is None:
                return interface_name, None

            return (
                Stripped(f"std::shared_ptr<{types_namespace}::{interface_name}>"),
                None,
            )

    elif isinstance(type_annotation, intermediate_type_inference.ListTypeAnnotation):
        item_type, error_msg = generate_type(
            type_annotation=type_annotation.items, types_namespace=types_namespace
        )
        if error_msg is not None:
            return None, error_msg

        assert item_type is not None
        if item_type.endswith(">"):
            return Stripped(f"std::vector<{item_type} >"), None

        return Stripped(f"std::vector<{item_type}>"), None

    elif isinstance(
        type_annotation, intermediate_type_inference.OptionalTypeAnnotation
    ):
        value_type, error_msg = generate_type(
            type_annotation=type_annotation.value, types_namespace=types_namespace
        )

        if error_msg is not None:
            return None, error_msg

        assert value_type is not None

        if value_type.endswith(">"):
            return Stripped(f"common::optional<{value_type} >"), None

        return Stripped(f"common::optional<{value_type}>"), None

    else:
        return None, (
            f"(mristin, 2023-06-01): We do not handle "
            f"the type annotation {type_annotation} from "
            "aas_core_codegen.intermediate.type_inference as that was, "
            "at this time point, YAGNI (*e.g.*, verification functions, "
            "built-in functions *etc.*). If you need this feature, please "
            "contact the developers."
        )

    raise AssertionError("Should not have gotten here")


def determine_whether_referencable(
    type_annotation: intermediate_type_inference.TypeAnnotationUnion,
) -> Tuple[Optional[bool], Optional[str]]:
    """
    Return ``True`` if the type annotation denotes a referencable value.

    (mristin, 2023-10-19): We do not handle all the type annotations from
    :py:mod:`aas_core_codegen.intermediate.type_inference` as that would be
    YAGNI (*e.g.*, verification functions, built-in functions *etc.*).
    If we do not know how to generate the type in C++, we return an error message.
    """
    if isinstance(type_annotation, intermediate_type_inference.PrimitiveTypeAnnotation):
        return False, None

    elif isinstance(type_annotation, intermediate_type_inference.OurTypeAnnotation):
        our_type = type_annotation.our_type

        if isinstance(our_type, intermediate.Enumeration):
            return False, None

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            return False, None

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            return True, None

    elif isinstance(type_annotation, intermediate_type_inference.ListTypeAnnotation):
        return True, None

    elif isinstance(
        type_annotation, intermediate_type_inference.OptionalTypeAnnotation
    ):
        return True, None

    else:
        return None, (
            f"(mristin, 2023-06-01): We do not handle "
            f"the type annotation {type_annotation} from "
            "aas_core_codegen.intermediate.type_inference as that was, "
            "at this time point, YAGNI (*e.g.*, verification functions, "
            "built-in functions *etc.*). If you need this feature, please "
            "contact the developers."
        )

    raise AssertionError("Should not have gotten here")


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def generate_type_with_const_ref_if_applicable(
    type_annotation: intermediate_type_inference.TypeAnnotationUnion,
    types_namespace: Optional[Identifier] = None,
) -> Tuple[Optional[Stripped], Optional[str]]:
    """
    Generate the C++ type and wrap it in ``const T&``, if applicable.

    If ``types_namespace`` is specified, it is prepended to all our types.

    (mristin, 2023-10-19): We do not handle all the type annotations from
    :py:mod:`aas_core_codegen.intermediate.type_inference` as that would be
    YAGNI (*e.g.*, verification functions, built-in functions *etc.*).
    If we do not know how to generate the type in C++, we return an error message.
    """
    code, error = generate_type(
        type_annotation=type_annotation, types_namespace=types_namespace
    )

    if error is not None:
        return None, error

    assert code is not None

    referencable, error = determine_whether_referencable(type_annotation)
    if error is not None:
        return None, error

    assert referencable is not None

    if referencable:
        return Stripped(f"const {code}&"), None

    return code, None


class Transpiler(
    parse_tree.RestrictedTransformer[Tuple[Optional[Stripped], Optional[Error]]]
):
    """Transpile a node of our AST to C++ code, or return an error."""

    _CPP_COMPARISON_MAP = {
        parse_tree.Comparator.LT: "<",
        parse_tree.Comparator.LE: "<=",
        parse_tree.Comparator.GT: ">",
        parse_tree.Comparator.GE: ">=",
        parse_tree.Comparator.EQ: "==",
        parse_tree.Comparator.NE: "!=",
    }

    def __init__(
        self,
        type_map: Mapping[
            parse_tree.Node, intermediate_type_inference.TypeAnnotationUnion
        ],
        is_optional_map: Mapping[parse_tree.Node, bool],
        environment: intermediate_type_inference.Environment,
        types_namespace: Optional[Identifier] = None,
    ) -> None:
        """
        Initialize with the given values.

        If ``types_namespace`` is specified, it is prepended to all our types.
        """
        self.type_map = type_map
        self.is_optional_map = is_optional_map
        self._environment = intermediate_type_inference.MutableEnvironment(
            parent=environment
        )
        self._types_namespace = types_namespace

        # NOTE (mristin, 2023-06-30):
        # Keep track whenever we define a variable name, so that we can know how to
        # resolve it as a name in the C++ code.
        #
        # While this class does not directly use it, the descendants of this class do!
        self._variable_name_set = set()  # type: Set[Identifier]

    def _transform_enumeration_literal(
        self, enumeration_name: Identifier, literal_name: Identifier
    ) -> Stripped:
        """Generate the code to represent an enumeration literal."""
        cpp_enum_name = cpp_naming.enum_name(enumeration_name)
        cpp_literal_name = cpp_naming.enum_literal_name(literal_name)
        if self._types_namespace is not None:
            return Stripped(
                f"{self._types_namespace}::{cpp_enum_name}::{cpp_literal_name}"
            )

        return Stripped(f"{cpp_enum_name}::{cpp_literal_name}")

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def _transform_and_value_if_necessary(
        self, node: parse_tree.Node
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        """
        Call ``.value()`` on the given node if it is a ``common::optional``.

        If the value denoted by ``node`` is not a ``common::optional``, it is returned
        transpiled as-is.
        """
        code, error = self.transform(node)
        if error is not None:
            return None, error

        if self.is_optional_map[node]:
            no_parentheses_types = (
                parse_tree.FunctionCall,
                parse_tree.Name,
                parse_tree.Constant,
            )
            if isinstance(node, no_parentheses_types):
                return Stripped(f"*{code}"), None

            return Stripped(f"(*({code}))"), None

        return code, None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_member(
        self, node: parse_tree.Member
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        instance, error = self._transform_and_value_if_necessary(node.instance)
        if error is not None:
            return None, error

        instance_type = self.type_map[node.instance]

        instance_type_beneath = intermediate_type_inference.beneath_optional(
            instance_type
        )

        member_type = self.type_map[node]
        member_type_beneath = intermediate_type_inference.beneath_optional(member_type)

        member_accessor: str

        if isinstance(
            instance_type_beneath, intermediate_type_inference.OurTypeAnnotation
        ) and isinstance(instance_type_beneath.our_type, intermediate.Enumeration):
            # NOTE (mristin, 2023-06-30):
            # This member denotes an enumeration literal of an enumeration.
            # In C++, enumeration literals are mere constants. Hence, we can not
            # "de-reference" the enumeration literals from an enumeration, but
            # generate the constant name here.
            return (
                self._transform_enumeration_literal(
                    enumeration_name=instance_type_beneath.our_type.name,
                    literal_name=node.name,
                ),
                None,
            )

        elif isinstance(
            member_type_beneath, intermediate_type_inference.MethodTypeAnnotation
        ):
            member_accessor = cpp_naming.method_name(node.name)

        elif isinstance(
            instance_type_beneath, intermediate_type_inference.OurTypeAnnotation
        ) and isinstance(instance_type_beneath.our_type, intermediate.Class):
            if node.name in instance_type_beneath.our_type.properties_by_name:
                getter_name = cpp_naming.getter_name(node.name)
                member_accessor = f"{getter_name}()"
            else:
                return None, Error(
                    node.original_node,
                    f"The property {node.name!r} has not been defined "
                    f"in the class {instance_type_beneath.our_type.name!r}",
                )

        elif isinstance(
            instance_type_beneath,
            intermediate_type_inference.EnumerationAsTypeTypeAnnotation,
        ):
            if node.name in instance_type_beneath.enumeration.literals_by_name:
                # NOTE (mristin, 2023-06-30):
                # The member denotes an enumeration literal of an enumeration.
                # In C++, enumeration literals are mere constants. Hence, we can not
                # "de-reference" the enumeration literals from an enumeration, but
                # generate the constant name here.
                return (
                    self._transform_enumeration_literal(
                        enumeration_name=instance_type_beneath.enumeration.name,
                        literal_name=node.name,
                    ),
                    None,
                )
            else:
                return None, Error(
                    node.original_node,
                    f"The literal {node.name!r} has not been defined "
                    f"in the enumeration {instance_type_beneath.enumeration.name!r}",
                )
        else:
            return None, Error(
                node.original_node,
                f"We do not know how to generate the member access. The inferred type "
                f"of the instance was {instance_type}, while the member type "
                f"was {member_type}. However, we do not know how to resolve "
                f"the member {node.name!r} in {instance_type}.",
            )

        assert isinstance(
            instance_type_beneath, intermediate_type_inference.OurTypeAnnotation
        ) and isinstance(instance_type_beneath.our_type, intermediate.Class), (
            f"The access to enumeration literals is expected to have been handled "
            f"before. If we got to this point, the instance is expected to be "
            f"an instance of a class, as we access its members with '->' and "
            f"assume it is a ``std::shared_ptr``. However, we got: {instance_type}"
        )

        return Stripped(f"{instance}->{member_accessor}"), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_index(
        self, node: parse_tree.Index
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        collection, error = self._transform_and_value_if_necessary(node.collection)
        if error is not None:
            return None, error

        index, error = self._transform_and_value_if_necessary(node.index)
        if error is not None:
            return None, error
        assert index is not None

        index_as_int = None  # type: Optional[int]
        try:
            index_as_int = int(index)
        except ValueError:
            pass

        if index_as_int is not None and index_as_int == -1:
            return Stripped(f"{collection}.back()"), None

        if index_as_int is not None and index_as_int < -1:
            # pylint: disable=invalid-unary-operand-type
            index = Stripped(f"{collection}.size() - {-index_as_int}")

        if "\n" in index:
            return (
                Stripped(
                    f"""\
{collection}.at(
{I}{indent_but_first_line(index, I)}
)"""
                ),
                None,
            )

        return Stripped(f"{collection}.at({index})"), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_comparison(
        self, node: parse_tree.Comparison
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        comparator = Transpiler._CPP_COMPARISON_MAP[node.op]

        errors = []

        left, error = self._transform_and_value_if_necessary(node.left)
        if error is not None:
            errors.append(error)

        right, error = self._transform_and_value_if_necessary(node.right)
        if error is not None:
            errors.append(error)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the comparison", errors
            )

        no_parentheses_types = (
            parse_tree.Member,
            parse_tree.FunctionCall,
            parse_tree.MethodCall,
            parse_tree.Name,
            parse_tree.Constant,
            parse_tree.Index,
        )

        if isinstance(node.left, no_parentheses_types) and isinstance(
            node.right, no_parentheses_types
        ):
            return Stripped(f"{left} {comparator} {right}"), None

        return Stripped(f"({left}) {comparator} ({right})"), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_is_in(
        self, node: parse_tree.IsIn
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []

        member, error = self._transform_and_value_if_necessary(node.member)
        if error is not None:
            errors.append(error)

        container, error = self._transform_and_value_if_necessary(node.container)
        if error is not None:
            errors.append(error)

        if len(errors) > 0:
            return None, Error(
                node.original_node,
                "Failed to transpile the membership relation",
                errors,
            )

        assert member is not None
        assert container is not None

        contains_function = cpp_naming.function_name(Identifier("contains"))

        return (
            Stripped(
                f"""\
common::{contains_function}(
{I}{indent_but_first_line(container, I)},
{I}{indent_but_first_line(member, I)}
)"""
            ),
            None,
        )

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_implication(
        self, node: parse_tree.Implication
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []

        antecedent, error = self._transform_and_value_if_necessary(node.antecedent)
        if error is not None:
            errors.append(error)

        consequent, error = self._transform_and_value_if_necessary(node.consequent)
        if error is not None:
            errors.append(error)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the implication", errors
            )

        assert antecedent is not None
        assert consequent is not None

        no_parentheses_types_in_this_context = (
            parse_tree.Member,
            parse_tree.FunctionCall,
            parse_tree.MethodCall,
            parse_tree.Name,
            parse_tree.IsIn,
            parse_tree.Index,
            parse_tree.All,
            parse_tree.Any,
        )

        if isinstance(node.antecedent, no_parentheses_types_in_this_context):
            not_antecedent = f"!{antecedent}"
        else:
            not_antecedent = f"!({antecedent})"

        if not isinstance(node.consequent, no_parentheses_types_in_this_context):
            consequent = Stripped(f"({consequent})")

        return Stripped(f"{not_antecedent}\n|| {consequent}"), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_method_call(
        self, node: parse_tree.MethodCall
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]

        member_access, error = self._transform_and_value_if_necessary(node.member)
        if error is not None:
            errors.append(error)

        args = []  # type: List[Stripped]
        for arg_node in node.args:
            arg_type = self.type_map[arg_node]

            # NOTE (mristin):
            # This is a tough call to make. We decide that a value, for which we
            # know that it might be null, will not be de-referenced. On the other
            # hand, if the value is certainly not null, we de-reference it.
            #
            # The problem here is that the actual type of the argument in C++ changes
            # depending on whether we check for its nullness before with an implication.

            arg: Optional[Stripped]
            if isinstance(arg_type, intermediate_type_inference.OptionalTypeAnnotation):
                arg, error = self.transform(arg_node)
            else:
                arg, error = self._transform_and_value_if_necessary(arg_node)

            if error is not None:
                errors.append(error)
                continue

            assert arg is not None

            args.append(arg)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the method call", errors
            )

        assert member_access is not None

        if len(args) == 0:
            return Stripped(f"{member_access}()"), None

        joined_args = ",\n".join(args)
        return (
            Stripped(
                f"""\
{member_access}(
{I}{indent_but_first_line(joined_args, I)}
)"""
            ),
            None,
        )

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_function_call(
        self, node: parse_tree.FunctionCall
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]

        args = []  # type: List[Stripped]
        for arg_node in node.args:
            arg_type = self.type_map[arg_node]

            # NOTE (mristin):
            # This is a tough call to make. We decide that a value, for which we
            # know that it might be null, will not be de-referenced. On the other
            # hand, if the value is certainly not null, we de-reference it.
            #
            # The problem here is that the actual type of the argument in C++ changes
            # depending on whether we check for its nullness before with an implication.
            arg: Optional[Stripped]
            if isinstance(arg_type, intermediate_type_inference.OptionalTypeAnnotation):
                arg, error = self.transform(arg_node)
            else:
                arg, error = self._transform_and_value_if_necessary(arg_node)

            if error is not None:
                errors.append(error)
                continue

            assert arg is not None

            args.append(arg)

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the function call", errors
            )

        # NOTE (mristin, 2023-06-30):
        # The validity of the arguments is checked in
        # :py:func:`aas_core_codegen.intermediate._translate.translate`, so we do not
        # have to test for argument arity here.

        func_type = self.type_map[node.name]

        if not isinstance(
            func_type, intermediate_type_inference.FunctionTypeAnnotationUnionAsTuple
        ):
            return None, Error(
                node.name.original_node,
                f"Expected the name to refer to a function, "
                f"but its inferred type was {func_type}",
            )

        if isinstance(
            func_type, intermediate_type_inference.VerificationTypeAnnotation
        ):
            function_name, error = self.transform_name(node.name)
            if error is not None:
                return None, error

            assert function_name is not None

            if len(args) == 0:
                return Stripped(f"{function_name}()"), None

            joined_args = ",\n".join(args)
            return (
                Stripped(
                    f"""\
{function_name}(
{I}{indent_but_first_line(joined_args, I)}
)"""
                ),
                None,
            )

        elif isinstance(
            func_type, intermediate_type_inference.BuiltinFunctionTypeAnnotation
        ):
            if func_type.func.name == "len":
                assert len(args) == 1, (
                    f"Expected exactly one argument, but got: {args}; "
                    f"this should have been caught before."
                )

                no_parentheses_types_in_this_context = (
                    parse_tree.Member,
                    parse_tree.FunctionCall,
                    parse_tree.MethodCall,
                    parse_tree.Name,
                    parse_tree.IsIn,
                    parse_tree.Index,
                    parse_tree.All,
                    parse_tree.Any,
                )

                first_arg, error = self._transform_and_value_if_necessary(node.args[0])
                if error is not None:
                    return None, error

                assert first_arg is not None

                if not isinstance(node.args[0], no_parentheses_types_in_this_context):
                    first_arg = Stripped(f"({first_arg})")

                return Stripped(f"{first_arg}.size()"), None

            else:
                return None, Error(
                    node.original_node,
                    f"The handling of the built-in function {node.name!r} has not "
                    f"been implemented",
                )
        else:
            assert_never(func_type)

        raise AssertionError("Should not have gotten here")

    def transform_constant(
        self, node: parse_tree.Constant
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if isinstance(node.value, bool):
            return Stripped("true" if node.value else "false"), None
        elif isinstance(node.value, (int, float)):
            return Stripped(cpp_common.float_literal(node.value)), None
        elif isinstance(node.value, str):
            return Stripped(cpp_common.wstring_literal(node.value)), None
        elif isinstance(node.value, bytes):
            literal, multiline = cpp_common.bytes_literal(node.value)

            if not multiline:
                return Stripped(literal), None
            else:
                return (
                    Stripped(
                        f"""\
(
{I}{indent_but_first_line(literal, I)}
)"""
                    ),
                    None,
                )
        else:
            assert_never(node.value)

        raise AssertionError("Should not have gotten here")

    def transform_is_none(
        self, node: parse_tree.IsNone
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        value, error = self.transform(node.value)
        if error is not None:
            return None, error

        no_parentheses_types = (
            parse_tree.Member,
            parse_tree.FunctionCall,
            parse_tree.MethodCall,
            parse_tree.Name,
            parse_tree.IsIn,
            parse_tree.Index,
            parse_tree.All,
            parse_tree.Any,
        )
        if isinstance(node.value, no_parentheses_types):
            return Stripped(f"!({value}.has_value())"), None
        else:
            return Stripped(f"!(({value}).has_value())"), None

    def transform_is_not_none(
        self, node: parse_tree.IsNotNone
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        value, error = self.transform(node.value)
        if error is not None:
            return None, error

        no_parentheses_types_in_this_context = (
            parse_tree.Member,
            parse_tree.FunctionCall,
            parse_tree.MethodCall,
            parse_tree.Name,
            parse_tree.IsIn,
            parse_tree.Index,
            parse_tree.All,
            parse_tree.Any,
        )
        if isinstance(node.value, no_parentheses_types_in_this_context):
            return Stripped(f"{value}.has_value()"), None
        else:
            return Stripped(f"({value}).has_value()"), None

    @abc.abstractmethod
    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        raise NotImplementedError()

    def transform_not(
        self, node: parse_tree.Not
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        operand, error = self._transform_and_value_if_necessary(node.operand)
        if error is not None:
            return None, error

        no_parentheses_types_in_this_context = (
            parse_tree.Member,
            parse_tree.FunctionCall,
            parse_tree.MethodCall,
            parse_tree.Name,
            parse_tree.IsIn,
            parse_tree.Index,
            parse_tree.All,
            parse_tree.Any,
        )
        if not isinstance(node.operand, no_parentheses_types_in_this_context):
            return Stripped(f"!({operand})"), None
        else:
            return Stripped(f"!{operand}"), None

    def _transform_and_or_or(
        self, node: Union[parse_tree.And, parse_tree.Or]
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]
        values = []  # type: List[Stripped]

        for value_node in node.values:
            value, error = self._transform_and_value_if_necessary(value_node)
            if error is not None:
                errors.append(error)
                continue

            assert value is not None

            no_parentheses_types_in_this_context = (
                parse_tree.Member,
                parse_tree.FunctionCall,
                parse_tree.MethodCall,
                parse_tree.Name,
                parse_tree.IsIn,
                parse_tree.Index,
                parse_tree.All,
                parse_tree.Any,
                parse_tree.Comparison,
            )

            if not isinstance(value_node, no_parentheses_types_in_this_context):
                # NOTE (mristin, 2023-06-30):
                # This is a very rudimentary heuristic for breaking the lines, and can
                # be greatly improved by rendering into C++ code. However, at this
                # point, we lack time for more sophisticated reformatting approaches.
                if "\n" in value:
                    value = Stripped(
                        f"""\
(
{I}{indent_but_first_line(value, I)}
)"""
                    )
                else:
                    value = Stripped(f"({value})")

            values.append(value)

        if len(errors) > 0:
            if isinstance(node, parse_tree.And):
                return None, Error(
                    node.original_node, "Failed to transpile the conjunction", errors
                )
            elif isinstance(node, parse_tree.Or):
                return None, Error(
                    node.original_node, "Failed to transpile the disjunction", errors
                )
            else:
                assert_never(node)

        assert len(values) >= 1
        if len(values) == 1:
            return Stripped(values[0]), None

        writer = io.StringIO()
        writer.write("(\n")
        for i, value in enumerate(values):
            if i == 0:
                writer.write(f"{I}{indent_but_first_line(value, I)}\n")
            else:
                if isinstance(node, parse_tree.And):
                    writer.write(f"{I}&& {indent_but_first_line(value, I)}\n")
                elif isinstance(node, parse_tree.Or):
                    writer.write(f"{I}|| {indent_but_first_line(value, I)}\n")
                else:
                    assert_never(node)

        writer.write(")")

        return Stripped(writer.getvalue()), None

    def transform_and(
        self, node: parse_tree.And
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        return self._transform_and_or_or(node)

    def transform_or(
        self, node: parse_tree.Or
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        return self._transform_and_or_or(node)

    def _transform_add_or_sub(
        self, node: Union[parse_tree.Add, parse_tree.Sub]
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]

        left, error = self._transform_and_value_if_necessary(node.left)
        if error is not None:
            errors.append(error)

        right, error = self._transform_and_value_if_necessary(node.right)
        if error is not None:
            errors.append(error)

        if len(errors) > 0:
            operation_name: str
            if isinstance(node, parse_tree.Add):
                operation_name = "the addition"
            elif isinstance(node, parse_tree.Sub):
                operation_name = "the subtraction"
            else:
                assert_never(node)

            return None, Error(
                node.original_node, f"Failed to transpile {operation_name}", errors
            )

        no_parentheses_types_in_this_context = (
            parse_tree.Member,
            parse_tree.FunctionCall,
            parse_tree.MethodCall,
            parse_tree.Name,
            parse_tree.IsIn,
            parse_tree.Index,
            parse_tree.All,
            parse_tree.Any,
        )

        if not isinstance(node.left, no_parentheses_types_in_this_context):
            left = Stripped(f"({left})")

        if not isinstance(node.right, no_parentheses_types_in_this_context):
            right = Stripped(f"({right})")

        if isinstance(node, parse_tree.Add):
            return Stripped(f"{left} + {right}"), None
        elif isinstance(node, parse_tree.Sub):
            return Stripped(f"{left} - {right}"), None
        else:
            assert_never(node)

    def transform_add(
        self, node: parse_tree.Add
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        return self._transform_add_or_sub(node)

    def transform_sub(
        self, node: parse_tree.Sub
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        return self._transform_add_or_sub(node)

    def transform_joined_str(
        self, node: parse_tree.JoinedStr
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if all(isinstance(value, str) for value in node.values):
            text = "".join(node.values)  # type: ignore
            return cpp_common.wstring_literal(text), None

        # NOTE (mristin, 2023-06-30):
        # We need the interpolation if we got so far.

        args = []  # type: List[str]

        for value in node.values:
            if isinstance(value, str):
                args.append(cpp_common.wstring_literal(value))

            elif isinstance(value, parse_tree.FormattedValue):
                code, error = self._transform_and_value_if_necessary(value.value)
                if error is not None:
                    return None, error

                assert code is not None

                value_type = self.type_map[value.value]

                if not isinstance(
                    value_type, intermediate_type_inference.PrimitiveTypeAnnotation
                ) and not (
                    isinstance(
                        value_type, intermediate_type_inference.OurTypeAnnotation
                    )
                    and isinstance(
                        value_type.our_type, intermediate.ConstrainedPrimitive
                    )
                ):
                    return None, Error(
                        value.original_node,
                        f"Unexpected non-primitive formatted value type: {value_type}",
                    )

                to_wstring = _determine_which_to_wstring(value_type)

                if to_wstring is None:
                    args.append(code)
                else:
                    if "\n" in code:
                        args.append(
                            f"""\
{to_wstring}(
{I}{indent_but_first_line(code, I)}
)"""
                        )
                    else:
                        args.append(f"{to_wstring}({code})")

        args_joined = ",\n".join(args)

        concat = cpp_naming.function_name(Identifier("concat"))
        return (
            Stripped(
                f"""\
common::{concat}(
{I}{indent_but_first_line(args_joined, I)}
)"""
            ),
            None,
        )

    def _transform_any_or_all(
        self, node: Union[parse_tree.Any, parse_tree.All]
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]

        iteration = None  # type: Optional[Stripped]
        start = None  # type: Optional[Stripped]
        end = None  # type: Optional[Stripped]

        if isinstance(node.generator, parse_tree.ForEach):
            iteration, error = self._transform_and_value_if_necessary(
                node.generator.iteration
            )
            if error is not None:
                errors.append(error)

        elif isinstance(node.generator, parse_tree.ForRange):
            start, error = self._transform_and_value_if_necessary(node.generator.start)
            if error is not None:
                errors.append(error)

            end, error = self._transform_and_value_if_necessary(node.generator.end)
            if error is not None:
                errors.append(error)

        else:
            assert_never(node.generator)

        variable_name = node.generator.variable.identifier
        variable_type_annotation = self.type_map[node.generator.variable]

        variable_name_cpp = cpp_naming.variable_name(variable_name)
        variable_type_cpp, error_msg = generate_type_with_const_ref_if_applicable(
            type_annotation=variable_type_annotation,
            types_namespace=self._types_namespace,
        )
        if error_msg is not None:
            errors.append(Error(node.generator.variable.original_node, error_msg))

        try:
            self._environment.set(
                identifier=variable_name, type_annotation=variable_type_annotation
            )
            self._variable_name_set.add(variable_name)

            condition, error = self._transform_and_value_if_necessary(node.condition)
            if error is not None:
                errors.append(error)

            variable, error = self._transform_and_value_if_necessary(
                node.generator.variable
            )
            if error is not None:
                errors.append(error)

        finally:
            self._variable_name_set.remove(variable_name)
            self._environment.remove(variable_name)

        if len(errors) > 0:
            return None, Error(
                node.original_node,
                "Failed to transpile the generator expression",
                errors,
            )

        assert (iteration is not None) ^ (start is not None and end is not None)

        assert variable is not None
        assert condition is not None

        if isinstance(node.generator, parse_tree.ForEach):
            assert iteration is not None

            qualifier_function: str
            if isinstance(node, parse_tree.Any):
                qualifier_function = cpp_naming.function_name(Identifier("Some"))
            elif isinstance(node, parse_tree.All):
                qualifier_function = cpp_naming.function_name(Identifier("All"))
            else:
                assert_never(node)

            no_parentheses_types_in_this_context = (
                parse_tree.Member,
                parse_tree.MethodCall,
                parse_tree.FunctionCall,
                parse_tree.Name,
                parse_tree.IsIn,
                parse_tree.Index,
                parse_tree.All,
                parse_tree.Any,
            )

            if not isinstance(
                node.generator.iteration, no_parentheses_types_in_this_context
            ):
                source = Stripped(f"({iteration})")
            else:
                source = iteration

            # NOTE (mristin, 2023-07-01):
            # We implicitly capture all the variables by reference,
            # see: https://en.cppreference.com/w/cpp/language/lambda#Lambda_capture.

            return (
                Stripped(
                    f"""\
common::{qualifier_function}(
{I}[&]({variable_type_cpp} {variable_name_cpp}) -> bool {{
{II}return {indent_but_first_line(condition, II)};
{I}}},
{I}{indent_but_first_line(source, I)}
)"""
                ),
                None,
            )

        elif isinstance(node.generator, parse_tree.ForRange):
            if isinstance(node, parse_tree.Any):
                qualifier_function = "SomeRange"
            elif isinstance(node, parse_tree.All):
                qualifier_function = "AllRange"
            else:
                assert_never(node)

            assert start is not None
            assert end is not None

            return (
                Stripped(
                    f"""\
common::{qualifier_function}(
{I}[&]({variable_type_cpp} {variable_name_cpp}) -> bool {{
{II}return {indent_but_first_line(condition, II)};
{I}}},
{I}{indent_but_first_line(start, I)},
{I}{indent_but_first_line(end, I)}
)"""
                ),
                None,
            )

        else:
            assert_never(node.generator)

    def transform_any(
        self, node: parse_tree.Any
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        return self._transform_any_or_all(node)

    def transform_all(
        self, node: parse_tree.All
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        return self._transform_any_or_all(node)

    def transform_assignment(
        self, node: parse_tree.Assignment
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        errors = []  # type: List[Error]

        value_type = self.type_map[node.value]

        is_definition = False

        target = None  # type: Optional[Stripped]
        if isinstance(node.target, parse_tree.Name):
            target_type = self._environment.find(identifier=node.target.identifier)
            if target_type is None:
                # NOTE (mristin, 2023-07-01):
                # This is a variable definition as we did not specify the identifier
                # in the environment.

                is_definition = True

                target_type = value_type
                self._variable_name_set.add(node.target.identifier)
                self._environment.set(
                    identifier=node.target.identifier, type_annotation=target_type
                )

                target, error = self.transform_name(node=node.target)
                if error is not None:
                    errors.append(error)
            else:
                target, error = self.transform(node=node.target)
                if error is not None:
                    errors.append(error)
        else:
            target_type = self.type_map[node.target]

        if len(errors) > 0:
            return None, Error(
                node.original_node, "Failed to transpile the assignment", errors
            )

        assert target is not None
        assert target_type is not None

        value: Optional[Stripped]

        if isinstance(
            target_type, intermediate_type_inference.OptionalTypeAnnotation
        ) and isinstance(
            value_type, intermediate_type_inference.OptionalTypeAnnotation
        ):
            value, error = self.transform(node.value)
        elif not isinstance(
            target_type, intermediate_type_inference.OptionalTypeAnnotation
        ) and isinstance(
            value_type, intermediate_type_inference.OptionalTypeAnnotation
        ):
            value, error = self._transform_and_value_if_necessary(node.value)
        else:
            # NOTE (mristin, 2023-07-01):
            # This is the case covering (target non-optional, value non-optional) and
            # (target optional, value non-optional).
            value, error = self.transform(node.value)

        if error is not None:
            return None, error
        assert value is not None

        maybe_definition_prefix = "" if not is_definition else "auto "

        # NOTE (mristin, 2022-07-12):
        # This is a rudimentary heuristic for basic line breaks, but works well in
        # practice.
        if "\n" in value or len(value) > 50:
            return (
                Stripped(
                    f"""\
{maybe_definition_prefix}{target} = (
{I}{indent_but_first_line(value, I)}
);"""
                ),
                None,
            )

        return Stripped(f"{maybe_definition_prefix}{target} = {value};"), None

    def transform_return(
        self, node: parse_tree.Return
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if node.value is None:
            return Stripped("return;"), None

        value, error = self.transform(node.value)
        if error is not None:
            return None, error

        assert value is not None

        # NOTE (mristin, 2023-06-30):
        # This is a rudimentary heuristic for basic line breaks, but works well in
        # practice.
        if "\n" in value or len(value) > 50:
            return (
                Stripped(
                    f"""\
return (
{I}{indent_but_first_line(value, I)}
);"""
                ),
                None,
            )

        return Stripped(f"return {value};"), None


# noinspection PyProtectedMember,PyProtectedMember
assert all(op in Transpiler._CPP_COMPARISON_MAP for op in parse_tree.Comparator)



"""Render the parsed (and possibly modified) regular expression back into values."""
import io
from typing import List, Union, Optional, Mapping

import more_itertools
from icontract import require, ensure

from aas_core_codegen.common import assert_never
from aas_core_codegen.parse.retree._types import (
    Transformer,
    Range,
    Regex,
    UnionExpr,
    Concatenation,
    Term,
    Symbol,
    Group,
    Char,
    Quantifier,
    CharSet,
    SymbolKind,
)
from aas_core_codegen.parse.tree import FormattedValue


class Renderer(Transformer[List[Union[str, FormattedValue]]]):
    """Render the regular expression back into a joined string."""

    _ESCAPING_IN_CHARACTER_LITERALS = {
        "\t": "\\t",
        "\n": "\\n",
        "\r": "\\r",
        "\f": "\\f",
        "\v": "\\v",
        ".": "\\.",
        "^": "\\^",
        "$": "\\$",
        "{": "\\{",
        "}": "\\}",
        "[": "\\[",
        "]": "\\]",
        "(": "\\(",
        ")": "\\)",
        "?": "\\?",
        "*": "\\*",
        "+": "\\+",
        "\\": "\\\\",
    }

    _ESCAPING_IN_RANGE = {
        "\t": "\\t",
        "\n": "\\n",
        "\r": "\\r",
        "\f": "\\f",
        "\v": "\\v",
        "[": "\\[",
        "]": "\\]",
        "\\": "\\\\",
        "-": "\\-",
    }

    @require(
        lambda escaping: all(len(key) == 1 for key in escaping),
        "The ``escaping`` works only on characters, not on arbitrary text",
    )
    # fmt: on
    def char_to_str_and_escape_or_encode_if_necessary(
        self, node: Char, escaping: Mapping[str, str]
    ) -> List[Union[str, FormattedValue]]:
        """Convert the ``node`` to a string, and escape and/or encode appropriately."""
        if not node.explicitly_encoded:
            escaped = escaping.get(node.character, None)
            if escaped is not None:
                result: List[Union[str, FormattedValue]] = [escaped]
            else:
                result = [node.character]

            return result
        else:
            code = ord(node.character)
            if code < 255:
                return [f"\\x{code:02x}"]
            else:
                return [node.character.encode("unicode_escape").decode("ascii")]

    def transform_regex(self, node: Regex) -> List[Union[str, FormattedValue]]:
        """Transform the ``regex``."""
        return self.transform(node.union)

    def transform_union_expr(self, node: UnionExpr) -> List[Union[str, FormattedValue]]:
        """Transform the ``union_expr``."""
        output = []  # type: List[Union[str, FormattedValue]]

        for i, concatenation in enumerate(node.uniates):
            if i > 0:
                output.append("|")

            output.extend(self.transform(concatenation))

        return output

    def transform_concatenation(
        self, node: Concatenation
    ) -> List[Union[str, FormattedValue]]:
        """Transform the ``concatenation``."""
        output = []  # type: List[Union[str, FormattedValue]]
        for concatenant in node.concatenants:
            output.extend(self.transform(concatenant))

        return output

    def transform_symbol(self, node: Symbol) -> List[Union[str, FormattedValue]]:
        """Transform the ``symbol``."""
        if node.kind is SymbolKind.START:
            return ["^"]
        elif node.kind is SymbolKind.END:
            return ["$"]
        elif node.kind is SymbolKind.DOT:
            return ["."]
        else:
            assert_never(node.kind)

    def transform_term(self, node: Term) -> List[Union[str, FormattedValue]]:
        """Transform the ``term``."""
        output = []  # type: List[Union[str, FormattedValue]]

        if isinstance(node.value, FormattedValue):
            output.append(node.value)
        elif isinstance(node.value, Char):
            output.extend(
                self.char_to_str_and_escape_or_encode_if_necessary(
                    node=node.value, escaping=Renderer._ESCAPING_IN_CHARACTER_LITERALS
                )
            )
        else:
            # noinspection PyTypeChecker
            output.extend(self.transform(node.value))

        if node.quantifier is not None:
            output.extend(self.transform(node.quantifier))

        return output

    def transform_group(self, node: Group) -> List[Union[str, FormattedValue]]:
        """Transform the ``group``."""
        output = ["("]  # type: List[Union[str, FormattedValue]]
        output.extend(self.transform(node.union))
        output.append(")")
        return output

    def transform_char(self, node: Char) -> List[Union[str, FormattedValue]]:
        """Transform the ``char``."""
        raise AssertionError(
            f"The responsibility of this method should have been covered either "
            f"by {Renderer.transform_term.__name__} "
            f"or {Renderer.transform_char_set.__name__}"
        )

    def transform_quantifier(
        self, node: Quantifier
    ) -> List[Union[str, FormattedValue]]:
        """Transform the ``quantifier``."""
        quantifier: str

        if node.maximum is not None:
            if node.minimum == node.maximum:
                quantifier = f"{{{node.minimum}}}"
            else:
                if node.minimum == 0:
                    if node.maximum == 1:
                        quantifier = "?"
                    else:
                        quantifier = f"{{0,{node.maximum}}}"
                else:
                    quantifier = f"{{{node.minimum},{node.maximum}}}"
        else:
            if node.minimum == 0:
                quantifier = "*"
            elif node.minimum == 1:
                quantifier = "+"
            else:
                quantifier = f"{{{node.minimum},}}"

        if node.non_greedy:
            quantifier += "?"

        return [quantifier]

    def transform_char_set(self, node: CharSet) -> List[Union[str, FormattedValue]]:
        """Transform the ``char_set``."""
        output = ["["]  # type: List[Union[str, FormattedValue]]

        already_output_something = False
        if node.complementing:
            already_output_something = True
            output.append("^")

        for i, a_range in enumerate(node.ranges):
            # NOTE (mristin, 2022-06-10):
            # The first and the last dash need no escaping.
            if (
                i in (0, len(node.ranges) - 1)
                and a_range.end is None
                and a_range.start.character == "-"
                and not a_range.start.explicitly_encoded
            ):
                output.append("-")

            # NOTE (mristin, 2022-06-10):
            # The caret needs to be escaped only if it is the very first character
            # in the character set.
            elif (
                i == 0
                and a_range.start.character == "^"
                and not a_range.start.explicitly_encoded
                and not already_output_something
            ):
                output.append("\\^")
            else:
                output.extend(
                    self.char_to_str_and_escape_or_encode_if_necessary(
                        node=a_range.start, escaping=Renderer._ESCAPING_IN_RANGE
                    )
                )

                if a_range.end is not None:
                    output.append("-")

                    output.extend(
                        self.char_to_str_and_escape_or_encode_if_necessary(
                            node=a_range.end, escaping=Renderer._ESCAPING_IN_RANGE
                        )
                    )

        output.append("]")

        return output

    def transform_range(self, node: Range) -> List[Union[str, FormattedValue]]:
        """Transform the ``range``."""
        raise AssertionError(
            f"The responsibility of this function should have been taken care of "
            f"by {Renderer.transform_char_set.__name__}"
        )


_RENDERER = Renderer()


# fmt: off
@ensure(
    lambda result:
    not (len(result) > 0)
    or (
        not any(
            isinstance(that, str) and isinstance(other, str)
            for that, other in more_itertools.windowed(result, 2)
        )
    )
)
# fmt: on
def render(
    regex: Regex, renderer: Optional[Renderer] = None
) -> List[Union[str, FormattedValue]]:
    """
    Render the regular expression ``regex`` to a joined string.

    If no ``renderer`` is supplied, the default renderer is used.
    """
    if renderer is None:
        renderer = _RENDERER

    values = renderer.transform(regex)

    # NOTE (mristin, 2022-06-10):
    # Compress the consecutive strings into a single string to have a shorter
    # representation

    output = []  # type: List[Union[str, FormattedValue]]
    buffer = io.StringIO()
    for value in values:
        if isinstance(value, str):
            buffer.write(value)
        elif isinstance(value, FormattedValue):
            buffered = buffer.getvalue()
            if len(buffered) > 0:
                output.append(buffered)
                buffer = io.StringIO()

            output.append(value)
        else:
            assert_never(value)

    buffered = buffer.getvalue()
    if len(buffered) > 0:
        output.append(buffered)

    return output



"""Generate ProtoBuf identifiers based on the identifiers from the meta-model."""

from typing import Union

from icontract import ensure

import aas_core_codegen.naming
from aas_core_codegen import intermediate
from aas_core_codegen.common import Identifier, assert_never


def interface_name(identifier: Identifier) -> Identifier:
    """
    Generate a ProtoBuf name for an interface based on its meta-model ``identifier``.

    This method is not to be used because proto3 does not support interfaces.
    """
    raise NotImplementedError("Interfaces are not supported by proto3.")


def enum_name(identifier: Identifier) -> Identifier:
    """
    Generate a ProtoBuf name for an enum based on its meta-model ``identifier``.

    >>> enum_name(Identifier("something"))
    'Something'

    >>> enum_name(Identifier("URL_to_something"))
    'UrlToSomething'
    """
    return aas_core_codegen.naming.capitalized_camel_case(identifier)


def enum_literal_name(identifier: Identifier) -> Identifier:
    """
    Generate a Protobuf name for an enum literal based on its meta-model ``identifier``.

    >>> enum_literal_name(Identifier("something"))
    'SOMETHING'

    >>> enum_literal_name(Identifier("URL_to_something"))
    'URL_TO_SOMETHING'
    """
    return aas_core_codegen.naming.upper_snake_case(identifier)


@ensure(
    lambda result: "_" not in result,
    "No underscode allowed so that we can attached our own suffixes such as ``_choice``",
)
def class_name(identifier: Identifier) -> Identifier:
    """
    Generate a ProtoBuf name for a class based on its meta-model ``identifier``.

    >>> class_name(Identifier("something"))
    'Something'

    >>> class_name(Identifier("URL_to_something"))
    'UrlToSomething'
    """
    return aas_core_codegen.naming.capitalized_camel_case(identifier)


def name_of(
    something: Union[
        intermediate.Enumeration, intermediate.ConcreteClass, intermediate.Interface
    ]
) -> Identifier:
    """Dispatch to the appropriate naming function."""
    if isinstance(something, intermediate.Enumeration):
        return enum_name(something.name)

    elif isinstance(something, intermediate.ConcreteClass):
        return class_name(something.name)

    elif isinstance(something, intermediate.Interface):
        return interface_name(something.name)

    else:
        assert_never(something)

    raise AssertionError("Should not have gotten here")


def property_name(identifier: Identifier) -> Identifier:
    """
    Generate a ProtoBuf name for a public property based on its meta-model ``identifier``.

    >>> property_name(Identifier("something"))
    'something'

    >>> property_name(Identifier("something_to_URL"))
    'something_to_url'
    """
    return aas_core_codegen.naming.lower_snake_case(identifier)


def private_property_name(identifier: Identifier) -> Identifier:
    """
    Generate a ProtoBuf name for a private property.

    This method is not to be used because proto3 does not support private properties.
    """
    raise NotImplementedError(
        "Private properties are not supported by proto3 Messages."
    )


def private_method_name(identifier: Identifier) -> Identifier:
    """
    Generate a ProtoBuf name for a private method.

    This method is not to be used because proto3 does not support private methods.
    """
    raise NotImplementedError("Methods are not supported by proto3 Messages.")


def method_name(identifier: Identifier) -> Identifier:
    """
    Generate a ProtoBuf name for a method.

    This method is not to be used because proto3 does not support methods.
    """
    raise NotImplementedError("Methods are not supported by proto3 Messages.")


def argument_name(identifier: Identifier) -> Identifier:
    """
    Generate a ProtoBuf name for an argument.

    This method is not to be used because proto3 does not support methods and
    thus no arguments.
    """
    raise NotImplementedError("Arguments are not supported by proto3 Messages.")


def variable_name(identifier: Identifier) -> Identifier:
    """
    Generate a ProtoBuf name for a variable.

    This method is not to be used because proto3 does not support variables.
    """
    raise NotImplementedError("Variables are not supported by proto3 Messages.")



"""Generate TypeScript code for JSON-ization based on the intermediate representation."""

import io
import textwrap
from typing import Tuple, Optional, List

from icontract import ensure

from aas_core_codegen import intermediate, naming, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    Identifier,
    assert_never,
    indent_but_first_line,
)
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
    description as typescript_description,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# region De-serialization


def _generate_bool_from_jsonable() -> Stripped:
    """Generate the function to decode a ``bool`` from a JSON-able."""
    return Stripped(
        f"""\
/**
 * Parse `jsonable` as a boolean.
 *
 * @param jsonable - to be parsed
 * @returns parsed boolean value, or an error
 */
function booleanFromJsonable(
{I}jsonable: JsonValue
): AasCommon.Either<boolean, DeserializationError> {{
{I}// `typeof` seems to be optimized these days, so we use it instead of
{I}// literal comparison, see:
{I}// https://stackoverflow.com/questions/61786250/is-typeof-faster-than-literal-comparison

{I}if (jsonable === null) {{
{II}return newDeserializationError<boolean>(
{III}"Expected a boolean, but got null"
{II});
{I}}}
{I}if (typeof jsonable !== "boolean") {{
{II}return newDeserializationError<boolean>(
{III}`Expected a boolean, but got ${{typeof jsonable}}`
{II});
{I}}}

{I}return new AasCommon.Either<boolean, DeserializationError>(jsonable, null);
}}"""
    )


def _generate_int_from_jsonable() -> Stripped:
    """Generate the function to decode an ``int`` from a JSON-able."""
    return Stripped(
        f"""\
/**
 * Parse `jsonable` as an integer.
 *
 * @param jsonable - to be parsed
 * @returns parsed integer value, or an error
 */
// eslint-disable-next-line @typescript-eslint/no-unused-vars
function integerFromJsonable(
{I}jsonable: JsonValue
): AasCommon.Either<number, DeserializationError> {{
{I}if (jsonable === null) {{
{II}return newDeserializationError<number>(
{III}"Expected an integer number, but got null"
{II});
{I}}}
{I}if (typeof jsonable !== "number") {{
{II}return newDeserializationError<number>(
{III}`Expected an integer number, but got: ${{typeof jsonable}}`
{II});
{I}}}

{I}if (!Number.isInteger(jsonable)) {{
{II}return newDeserializationError<number>(
{III}`Expected an integer number, but got: ${{jsonable}}`
{II});
{I}}}

{I}return new AasCommon.Either<number, DeserializationError>(jsonable, null);
}}"""
    )


def _generate_float_from_jsonable() -> Stripped:
    """Generate the function to decode a ``float`` from a JSON-able."""
    return Stripped(
        f"""\
/**
 * Parse `jsonable` as a number.
 *
 * @param jsonable - to be parsed
 * @returns parsed numeric value, or an error
 */
// eslint-disable-next-line @typescript-eslint/no-unused-vars
function numberFromJsonable(
{I}jsonable: JsonValue
): AasCommon.Either<number, DeserializationError> {{
{I}if (jsonable === null) {{
{II}return newDeserializationError<number>(
{III}"Expected a number, but got null"
{II});
{I}}}
{I}if (typeof jsonable !== "number") {{
{II}return newDeserializationError<number>(
{III}`Expected a number, but got: ${{typeof jsonable}}`
{II});
{I}}}

{I}return new AasCommon.Either<number, DeserializationError>(jsonable, null);
}}"""
    )


def _generate_str_from_jsonable() -> Stripped:
    """Generate the function to decode a ``str`` from a JSON-able."""
    return Stripped(
        f"""\
/**
 * Parse `jsonable` as a string.
 *
 * @param jsonable - to be parsed
 * @returns parsed string value, or an error
 */
function stringFromJsonable(
{I}jsonable: JsonValue
): AasCommon.Either<string, DeserializationError> {{
{I}if (jsonable === null) {{
{II}return newDeserializationError<string>(
{III}"Expected a string, but got null"
{II});
{I}}}
{I}if (typeof jsonable !== "string") {{
{II}return newDeserializationError<string>(
{III}`Expected a string, but got: ${{typeof jsonable}}`
{II});
{I}}}

{I}return new AasCommon.Either<string, DeserializationError>(jsonable, null);
}}"""
    )


def _generate_bytes_from_jsonable() -> Stripped:
    """Generate the function to decode ``bytes`` from a JSON-able."""
    return Stripped(
        f"""\
/**
 * Parse `jsonable` as a byte array.
 *
 * @param jsonable - to be parsed
 * @returns parsed byte array, or an error
 */
function bytesFromJsonable(
{I}jsonable: JsonValue
): AasCommon.Either<Uint8Array, DeserializationError> {{
{I}if (jsonable === null) {{
{II}return newDeserializationError<Uint8Array>(
{III}"Expected a base64-encoded string, but got null"
{II});
{I}}}
{I}if (typeof jsonable !== "string") {{
{II}return newDeserializationError<Uint8Array>(
{III}`Expected a base64-encoded string, but got: ${{typeof jsonable}}`
{II});
{I}}}

{I}const either = AasCommon.base64Decode(jsonable);
{I}if (either.error !== null) {{
{II}return newDeserializationError<Uint8Array>(either.error);
{I}}}
{I}return new AasCommon.Either<Uint8Array, DeserializationError>(
{II}either.mustValue(), null
{I});
}}"""
    )


def _generate_enumeration_from_jsonable(
    enumeration: intermediate.Enumeration,
) -> Stripped:
    """Generate the deserialization method for an enumeration."""
    enum_name = typescript_naming.enum_name(identifier=enumeration.name)

    function_name = typescript_naming.function_name(
        Identifier(f"{enumeration.name}_from_jsonable")
    )

    enum_from_str = typescript_naming.function_name(
        Identifier(f"{enumeration.name}_from_string")
    )

    return Stripped(
        f"""\
/**
 * Parse `jsonable` structure as a literal
 * of {{@link {typescript_common.TYPES_MODULE}!{enum_name}}}.
 *
 * @param jsonable - to be parsed
 * @returns parsed literal, or an error if `jsonable` invalid
 */
export function {function_name}(
{I}jsonable: JsonValue
): AasCommon.Either<AasTypes.{enum_name}, DeserializationError> {{
{I}if (typeof jsonable !== "string") {{
{II}return newDeserializationError<AasTypes.{enum_name}>(
{III}`Expected a string, but got: ${{typeof jsonable}}`
{II});
{I}}}

{I}const literal = AasStringification.{enum_from_str}(jsonable);
{I}if (literal === null) {{
{II}return newDeserializationError<AasTypes.{enum_name}>(
{III}"Not a valid string representation of " +
{IIII}`a literal of {enum_name}: ${{jsonable}}`
{II});
{I}}}

{I}return new AasCommon.Either<
{II}AasTypes.{enum_name},
{II}DeserializationError
{I}>(literal, null);
}}"""
    )


def _generate_dispatch_map_for_interface(
    interface: intermediate.Interface,
) -> Stripped:
    """Generate a mapping model type 🠒 de-serialization function."""
    assert len(interface.base.concrete_descendants) > 0, (
        "Expected a class to have concrete descendants. "
        "Otherwise we do not know how to de-serialize it."
    )

    mapping_name = typescript_naming.constant_name(
        Identifier(f"{interface.name}_from_jsonable_dispatch")
    )

    interface_name = typescript_naming.interface_name(interface.name)

    mapping_writer = io.StringIO()
    mapping_writer.write(
        f"""\
const {mapping_name} =
{I}new Map<
{II}string,
{II}(JsonValue) => AasCommon.Either<
{III}AasTypes.{interface_name},
{III}DeserializationError
{II}>
{I}>(
{II}[
"""
    )

    for i, implementer in enumerate(interface.implementers):
        if len(implementer.concrete_descendants) == 0:
            function_name_for_implementer = typescript_naming.function_name(
                Identifier(f"{implementer.name}_from_jsonable")
            )
        else:
            # NOTE (mristin, 2022-11-25):
            # We can not use the public function as it would end in an endless dispatch
            # loop. Hence, we introduce a function which assumes the type and explicitly
            # does not dispatch.
            function_name_for_implementer = typescript_naming.function_name(
                Identifier(f"{implementer.name}_from_jsonable_without_dispatch")
            )

        implementer_literal = typescript_common.string_literal(
            naming.json_model_type(implementer.name)
        )

        mapping_writer.write(
            f"""\
{III}[
{IIII}{implementer_literal},
{IIII}{function_name_for_implementer}
{III}]"""
        )

        if i < len(interface.implementers) - 1:
            mapping_writer.write(",\n")
        else:
            mapping_writer.write("\n")

    mapping_writer.write(
        f"""\
{II}]
{I});"""
    )

    return Stripped(mapping_writer.getvalue())


def _generate_dispatch_from_jsonable(interface: intermediate.Interface) -> Stripped:
    """Generate the de-serialization dispatch for an abstract class."""
    function_name = typescript_naming.function_name(
        Identifier(f"{interface.name}_from_jsonable")
    )

    interface_name = typescript_naming.interface_name(interface.name)

    mapping_name = typescript_naming.constant_name(
        Identifier(f"{interface.name}_from_jsonable_dispatch")
    )

    return Stripped(
        f"""\
/**
 * Parse `jsonable` as an instance
 * of {{@link {typescript_common.TYPES_MODULE}!{interface_name}}}.
 *
 * @param jsonable - to be parsed
 * @returns parsed instance, or error if `jsonable` is invalid
 */
export function {function_name}(
{I}jsonable: JsonValue
): AasCommon.Either<
{I}AasTypes.{interface_name},
{I}DeserializationError
> {{
{I}if (jsonable === null) {{
{II}return newDeserializationError<AasTypes.{interface_name}>(
{III}"Expected a JSON object, but got null"
{II});
{I}}}
{I}if (Array.isArray(jsonable)) {{
{II}return newDeserializationError<AasTypes.{interface_name}>(
{III}"Expected a JSON object, but got a JSON array"
{II});
{I}}}
{I}if (typeof jsonable !== "object") {{
{II}return newDeserializationError<AasTypes.{interface_name}>(
{III}`Expected a JSON object, but got: ${{typeof jsonable}}`
{II});
{I}}}

{I}const modelType = jsonable["modelType"];
{I}if (modelType === undefined) {{
{II}return newDeserializationError<AasTypes.{interface_name}>(
{III}"The required property modelType is missing"
{II});
{I}}}

{I}if (typeof modelType !== "string") {{
{II}return newDeserializationError<AasTypes.{interface_name}>(
{III}`Expected the property modelType to be a string, but got: ${{typeof modelType}}`
{II});
{I}}}

{I}const dispatch = {mapping_name}.get(modelType);
{I}if (dispatch === undefined) {{
{II}return newDeserializationError<AasTypes.{interface_name}>(
{III}`Unexpected model type for {interface_name}: ${{modelType}}`
{II});
{I}}}

{I}return dispatch(jsonable);
}}"""
    )


_PARSE_FUNCTION_BY_PRIMITIVE_TYPE = {
    intermediate.PrimitiveType.BOOL: "booleanFromJsonable",
    intermediate.PrimitiveType.INT: "integerFromJsonable",
    intermediate.PrimitiveType.FLOAT: "numberFromJsonable",
    intermediate.PrimitiveType.STR: "stringFromJsonable",
    intermediate.PrimitiveType.BYTEARRAY: "bytesFromJsonable",
}
assert all(
    literal in _PARSE_FUNCTION_BY_PRIMITIVE_TYPE
    for literal in intermediate.PrimitiveType
)


def _parse_function_for_atomic_value(
    type_annotation: intermediate.AtomicTypeAnnotation,
) -> Stripped:
    """Determine the parse function for deserializing an atomic non-optional value."""
    function_name: str

    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        function_name = _PARSE_FUNCTION_BY_PRIMITIVE_TYPE[type_annotation.a_type]

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        our_type = type_annotation.our_type

        if isinstance(our_type, intermediate.Enumeration):
            function_name = typescript_naming.function_name(
                Identifier(f"{our_type.name}_from_jsonable")
            )

        elif isinstance(
            our_type,
            (
                intermediate.AbstractClass,
                intermediate.ConcreteClass,
            ),
        ):
            if our_type.interface is not None:
                assert our_type.interface.name == our_type.name, (
                    "Assume that the interface name and the class name in "
                    "the intermediate representation are the same, so that the "
                    "``*_from_jsonable`` name makes sense in all cases"
                )

            function_name = typescript_naming.function_name(
                Identifier(f"{our_type.name}_from_jsonable")
            )

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            function_name = _PARSE_FUNCTION_BY_PRIMITIVE_TYPE[our_type.constrainee]

        else:
            assert_never(our_type)
    else:
        assert_never(type_annotation)

    return Stripped(function_name)


def _generate_setter(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate a class which allows us to dispatch de-serialization for properties."""
    blocks = []  # type: List[Stripped]

    for i, prop in enumerate(cls.properties):
        prop_name = typescript_naming.property_name(prop.name)
        prop_type = typescript_common.generate_type(
            prop.type_annotation, types_module=Identifier("AasTypes")
        )

        # NOTE (mristin, 2022-11-25):
        # We make all the properties optional since we switch over the properties
        # during the de-serialization.
        if not isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            prop_type = Stripped(f"{prop_type} | null")

        blocks.append(Stripped(f"{prop_name}: {prop_type} = null;"))

    if cls.serialization.with_model_type:
        # NOTE (mristin):
        # If the serialization requires a model type, we consequently parse and set it
        # in the setter. The model type thus obtained is *not* used for any dispatch. We
        # only use this value for verification to make sure that the model type
        # of the instances is consistent with the expected value for its concrete
        # class. This will be performed even though the code might have had to parse
        # model type before for the dispatch. We decided to double-check to cover the
        # case where a dispatch is *unnecessary* (*e.g.*, the caller knows the expected
        # runtime type), but the model type might still be invalid in the input. Hence,
        # when the dispatch is *necessary*, the model type JSON property will be parsed
        # twice, which is a cost we currently find acceptable.
        prop_name = typescript_naming.property_name(Identifier("model_type"))
        blocks.append(
            Stripped(
                f"""\
// Used only for verification, not for dispatch!
{prop_name}: string | null = null;"""
            )
        )

    for i, prop in enumerate(cls.properties):
        prop_name = typescript_naming.property_name(prop.name)

        type_anno = intermediate.beneath_optional(prop.type_annotation)
        if isinstance(
            type_anno,
            (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
        ):
            function_name = _parse_function_for_atomic_value(type_anno)
            body = Stripped(
                f"""\
const parsedOrError = {function_name}(
{I}jsonable
);
if (parsedOrError.error !== null) {{
{I}return parsedOrError.error;
}} else {{
{I}this.{prop_name} = parsedOrError.mustValue();
{I}return null;
}}"""
            )

        elif isinstance(type_anno, intermediate.ListTypeAnnotation):
            assert not isinstance(
                type_anno.items,
                (intermediate.OptionalTypeAnnotation, intermediate.ListTypeAnnotation),
            ), (
                "We chose to implement only a very limited pattern matching; "
                "see intermediate._translate_._verify_only_simple_type_patterns"
            )

            items_type = typescript_common.generate_type(
                type_anno.items, types_module=Identifier("AasTypes")
            )
            parse_function = _parse_function_for_atomic_value(type_anno.items)

            body = Stripped(
                f"""\
if (jsonable === null) {{
{I}return new DeserializationError(
{II}"Expected an iterable, but got null"
{I});
}}
if (typeof jsonable !== "object") {{
{I}return new DeserializationError(
{II}`Expected an iterable, but got: ${{typeof jsonable}}`
{I});
}}
if (typeof jsonable[Symbol.iterator] !== "function") {{
{I}return new DeserializationError(
{II}"Expected an iterable with iterator function, " +
{III}`but got iterator of type: ${{typeof jsonable[Symbol.iterator]}}`
{I});
}}

const iterable = <Iterable<JsonValue>>jsonable;

const items =
{I}new Array<{items_type}>();

let i = 0;
for (const jsonableItem of iterable) {{
{I}const itemOrError = {parse_function}(
{II}jsonableItem
{I});

{I}if (itemOrError.error !== null) {{
{II}itemOrError.error.path.prepend(
{III}new IndexSegment(
{IIII}iterable,
{IIII}i
{III})
{II});
{II}return itemOrError.error;
{I}}}

{I}items.push(itemOrError.mustValue());
{I}i++;
}}

this.{prop_name} = items;
return null;"""
            )

        else:
            assert_never(type_anno)

        method_name = typescript_naming.method_name(
            Identifier(f"set_{prop.name}_from_jsonable")
        )

        method_writer = io.StringIO()
        method_writer.write(
            f"""\
/**
 * Parse `jsonable` as the value of {{@link {prop_name}}}.
 *
 * @param jsonable - to be parsed
 * @returns error, if any
 */
{method_name}(
{I}jsonable: JsonValue
): DeserializationError | null {{
{I}{indent_but_first_line(body, I)}
}}"""
        )

        blocks.append(Stripped(method_writer.getvalue()))

    if cls.serialization.with_model_type:
        method_name = typescript_naming.method_name(
            Identifier("set_model_type_from_jsonable")
        )
        prop_name = typescript_naming.property_name(Identifier("model_type"))

        blocks.append(
            Stripped(
                f"""\
/**
 * Parse `jsonable` as the model type of the concrete instance.
 *
 * This is intended only for verification, and no dispatch is performed.
 *
 * @param jsonable - to be parsed
 * @returns error, if any
 */
{method_name}(
{I}jsonable: JsonValue
): DeserializationError | null {{
{I}const parsedOrError = stringFromJsonable(
{II}jsonable
{I});
{I}if (parsedOrError.error !== null) {{
{II}return parsedOrError.error;
{I}}} else {{
{II}this.{prop_name} = parsedOrError.mustValue();
{II}return null;
{I}}}
}}"""
            )
        )

    cls_name = typescript_naming.class_name(cls.name)
    setter_cls_name = typescript_naming.class_name(Identifier(f"Setter_for_{cls.name}"))

    writer = io.StringIO()
    writer.write(
        f"""\
/**
 * Provide de-serialize & set methods for properties
 * of {{@link {typescript_common.TYPES_MODULE}!{cls_name}}}.
 */
class {setter_cls_name} {{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_setter_map(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate a map ``JSON property name`` -> deserialization on a setter."""
    # fmt: off
    assert (
            sorted(
                (arg.name, str(arg.type_annotation))
                for arg in cls.constructor.arguments
            ) == sorted(
                (prop.name, str(prop.type_annotation))
                for prop in cls.properties
            )
    ), (
        "(mristin, 2022-11-25) We assume that the properties and constructor arguments "
        "are identical at this point. If this is not the case, we have to re-write the "
        "logic substantially! Please contact the developers if you see this."
    )
    # fmt: on

    identifiers_expressions = []  # type: List[Tuple[Identifier, Stripped]]

    setter_cls_name = typescript_naming.class_name(Identifier(f"Setter_for_{cls.name}"))

    for prop in cls.properties:
        json_identifier = naming.json_property(prop.name)
        method_name = typescript_naming.method_name(
            Identifier(f"set_{prop.name}_from_jsonable")
        )

        identifiers_expressions.append(
            (json_identifier, Stripped(f"{setter_cls_name}.prototype.{method_name}"))
        )

    map_name = typescript_naming.constant_name(Identifier(f"setter_map_for_{cls.name}"))

    writer = io.StringIO()
    writer.write(
        f"""\
const {map_name} =
{I}new Map<
{II}string,
{II}(
{III}jsonable: JsonValue
{II}) => DeserializationError | null
{I}>(
{II}[
"""
    )

    for identifier, expression in identifiers_expressions:
        writer.write(
            f"""\
{III}[
{IIII}{typescript_common.string_literal(identifier)},
{IIII}{indent_but_first_line(expression, IIII)}
{III}],
"""
        )

    if cls.serialization.with_model_type:
        # NOTE (mristin):
        # If the serialization requires a model type, we consequently parse and set it
        # in the setter. The model type thus obtained is *not* used for any dispatch. We
        # only use this value for verification to make sure that the model type
        # of the instances is consistent with the expected value for its concrete
        # class. This will be performed even though the code might have had to parse
        # model type before for the dispatch. We decided to double-check to cover the
        # case where a dispatch is *unnecessary* (*e.g.*, the caller knows the expected
        # runtime type), but the model type might still be invalid in the input. Hence,
        # when the dispatch is *necessary*, the model type JSON property will be parsed
        # twice, which is a cost we currently find acceptable.
        json_identifier = naming.json_property(Identifier("model_type"))
        method_name = typescript_naming.method_name(
            Identifier("set_model_type_from_jsonable")
        )
        expression = Stripped(f"{setter_cls_name}.prototype.{method_name}")

        writer.write(
            f"""\
{III}[
{IIII}// The model type here is used only for verification, not for dispatch.
{IIII}{typescript_common.string_literal(json_identifier)},
{IIII}{indent_but_first_line(expression, IIII)}
{III}],
"""
        )

    writer.write(
        f"""\
{II}]
{I});"""
    )

    return Stripped(writer.getvalue())


def _generate_concrete_class_from_jsonable(
    cls: intermediate.ConcreteClass,
) -> Stripped:
    """
    Generate the deserialization function for a concrete class.

    This function performs no dispatch. If it de-serializes a concrete class with
    concrete descendants, we have to provide a different name. Otherwise, it would
    shadow the name for the dispatch function.
    """
    # fmt: off
    assert (
            sorted(
                (arg.name, str(arg.type_annotation))
                for arg in cls.constructor.arguments
            ) == sorted(
                (prop.name, str(prop.type_annotation))
                for prop in cls.properties
            )
    ), (
        "(mristin, 2022-11-30) We assume that the properties and constructor arguments "
        "are identical at this point. If this is not the case, we have to re-write the "
        "logic substantially! Please contact the developers if you see this."
    )
    # fmt: on

    cls_name = typescript_naming.class_name(cls.name)

    setter_cls_name = typescript_naming.class_name(Identifier(f"Setter_for_{cls.name}"))

    blocks = [
        Stripped(
            f"""\
if (jsonable === null) {{
{I}return newDeserializationError<AasTypes.{cls_name}>(
{II}"Expected a JSON object, but got null"
{I});
}}
{I}if (Array.isArray(jsonable)) {{
{II}return newDeserializationError<AasTypes.{cls_name}>(
{III}"Expected a JSON object, but got a JSON array"
{II});
{I}}}
if (typeof jsonable !== "object") {{
{I}return newDeserializationError<AasTypes.{cls_name}>(
{II}`Expected a JSON object, but got: ${{typeof jsonable}}`
{I});
}}"""
        ),
        Stripped(f"const setter = new {setter_cls_name}();"),
    ]  # type: List[Stripped]

    # region Switch on property name

    map_name = typescript_naming.constant_name(Identifier(f"setter_map_for_{cls.name}"))

    blocks.append(
        Stripped(
            f"""\
for (const key in jsonable) {{
{I}const jsonableValue = jsonable[key];
{I}const setterMethod =
{II}{map_name}.get(key);

{I}// NOTE (mristin, 2022-11-30):
{I}// Since we conflate here a JavaScript object with a JSON object, we ignore
{I}// properties which we do not know how to de-serialize and assume they are
{I}// related to the *JavaScript* properties of the object or `Object` prototype.
{I}if (setterMethod === undefined) {{
{II}continue;
{I}}}

{I}const error = setterMethod.call(setter, jsonableValue);
{I}if (error !== null) {{
{II}error.path.prepend(
{III}new PropertySegment(<JsonObject>jsonable, key)
{II});
{II}return new AasCommon.Either<
{III}AasTypes.{cls_name},
{III}DeserializationError
{II}>(
{IIII}null,
{IIII}error
{III});
{I}}}
}}"""
        )
    )

    # region Check required properties

    required_checks = []  # type: List[Stripped]
    for i, prop in enumerate(cls.properties):
        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            continue

        prop_name = typescript_naming.property_name(prop.name)

        message_literal = typescript_common.string_literal(
            f"The required property {naming.json_property(prop.name)!r} is missing"
        )
        required_checks.append(
            Stripped(
                f"""\
if (setter.{prop_name} === null) {{
{I}return newDeserializationError<
{II}AasTypes.{cls_name}
{I}>(
{II}{message_literal}
{I});
}}"""
            )
        )

    if len(required_checks) > 0:
        blocks.append(Stripped("\n\n".join(required_checks)))

    # endregion

    if cls.serialization.with_model_type:
        model_type = naming.json_model_type(cls.name)
        prop_name = typescript_naming.property_name(Identifier("model_type"))

        blocks.append(
            Stripped(
                f"""\
if (setter.{prop_name} === null) {{
{I}return newDeserializationError<
{II}AasTypes.{cls_name}
{I}>(
{II}"The required property 'modelType' is missing"
{I});
}} else if (setter.{prop_name} != "{model_type}") {{
{I}return newDeserializationError<
{II}AasTypes.{cls_name}
{I}>(
{II}"Expected model type '{model_type}', " +
{II}`but got: ${{setter.{prop_name}}}`
{I});
}}"""
            )
        )

    # region Pass in arguments to the constructor

    cls_name = typescript_naming.class_name(cls.name)

    if len(cls.constructor.arguments) == 0:
        blocks.append(
            Stripped(
                f"""\
return new AasCommon.Either<
{I}AasTypes.{cls_name},
{I}DeserializationError
>(
{I}new AasTypes.{cls_name}(),
{I}null
);"""
            )
        )
    else:
        init_writer = io.StringIO()
        init_writer.write(
            f"""\
return new AasCommon.Either<
{I}AasTypes.{cls_name},
{I}DeserializationError
>(
{I}new AasTypes.{cls_name}(
"""
        )

        for i, arg in enumerate(cls.constructor.arguments):
            prop = cls.properties_by_name[arg.name]

            prop_name = typescript_naming.property_name(prop.name)

            init_writer.write(f"{II}setter.{prop_name}")

            if i < len(cls.constructor.arguments) - 1:
                init_writer.write(",\n")
            else:
                init_writer.write("\n")

        init_writer.write(
            f"""\
{I}),
{I}null
);"""
        )

        blocks.append(Stripped(init_writer.getvalue()))
    # endregion

    writer = io.StringIO()

    if len(cls._concrete_descendants) == 0:
        function_name = typescript_naming.function_name(
            Identifier(f"{cls.name}_from_jsonable")
        )
    else:
        function_name = typescript_naming.function_name(
            Identifier(f"{cls.name}_from_jsonable_without_dispatch")
        )

    description_blocks = [
        Stripped(
            f"""\
Parse an instance of {{@link {typescript_common.TYPES_MODULE}!{cls_name}}} from the JSON-able
structure `jsonable`."""
        )
    ]  # type: List[Stripped]

    if len(cls.concrete_descendants) > 0:
        function_name_with_dispatch = typescript_naming.function_name(
            Identifier(f"{cls.name}_from_jsonable")
        )

        description_blocks.append(
            Stripped(
                f"""\
This function performs no dispatch! It is used to parse the properties
as-are, and already assumes the exact model type. Usually, this function
is called from within a dispatching function, and you never call it
directly. If you want to de-serialize an instance of
{{@link {typescript_common.TYPES_MODULE}!{cls_name}}}, call
{{@link {function_name_with_dispatch}}}."""
            )
        )

    description_blocks.append(
        Stripped(
            f"""\
@param jsonable - structure to be parsed
@returns parsed instance of {{@link {typescript_common.TYPES_MODULE}!{cls_name}}},
or an error if any"""
        )
    )

    description = "\n\n".join(description_blocks)
    description_comment = typescript_description.documentation_comment(
        Stripped(description)
    )

    # NOTE (mristin, 2022-11-30):
    # We export it only if the de-serialization of the class is equivalent to
    # the de-serialization without a dispatch.
    maybe_export_prefix = "export " if len(cls.concrete_descendants) == 0 else ""

    writer.write(
        f"""\
{description_comment}
{maybe_export_prefix}function {function_name}(
{I}jsonable: JsonValue
): AasCommon.Either<
{I}AasTypes.{cls_name},
{I}DeserializationError
> {{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


# endregion

# region Serialization


def _generate_transform_atomic_value(
    access_expression: str, type_anno: intermediate.AtomicTypeAnnotation
) -> Stripped:
    """
    Generate the snippet to transform the ``access_expression``.

    The ``access_expression`` should either be a name or a member access.
    """
    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation) or (
        isinstance(type_anno, intermediate.OurTypeAnnotation)
        and isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive)
    ):
        a_type = intermediate.try_primitive_type(type_anno)
        assert a_type is not None

        if (
            a_type is intermediate.PrimitiveType.BOOL
            or a_type is intermediate.PrimitiveType.INT
            or a_type is intermediate.PrimitiveType.FLOAT
            or a_type is intermediate.PrimitiveType.STR
        ):
            return Stripped(f"{access_expression}")

        elif a_type is intermediate.PrimitiveType.BYTEARRAY:
            return Stripped(f"AasCommon.base64Encode({access_expression})")

        else:
            assert_never(a_type)

    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        if isinstance(type_anno.our_type, intermediate.Enumeration):
            must_to_str_name = typescript_naming.function_name(
                Identifier(f"must_{type_anno.our_type.name}_to_string")
            )

            return Stripped(
                f"""\
AasStringification.{must_to_str_name}(
{I}{indent_but_first_line(access_expression, I)}
)"""
            )

        elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
            raise AssertionError("This case should have been handled before.")

        elif isinstance(
            type_anno.our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            return Stripped(f"this.transform({access_expression})")

        else:
            assert_never(type_anno.our_type)

    else:
        assert_never(type_anno.our_type)


def _generate_transform(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the ``transformX`` method to serialize an instance into a JSON-able."""
    blocks = [Stripped("const jsonable: JsonObject = {};")]  # type: List[Stripped]

    for prop in cls.properties:
        key_literal = typescript_common.string_literal(naming.json_property(prop.name))
        prop_name = typescript_naming.property_name(prop.name)

        type_anno = intermediate.beneath_optional(prop.type_annotation)

        block: Stripped

        if isinstance(
            type_anno,
            (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
        ):
            transformation_expression = _generate_transform_atomic_value(
                access_expression=Stripped(f"that.{prop_name}"), type_anno=type_anno
            )

            block = Stripped(
                f"""\
jsonable[{key_literal}] =
{I}{indent_but_first_line(transformation_expression, I)};"""
            )

        elif isinstance(type_anno, intermediate.ListTypeAnnotation):
            assert isinstance(
                type_anno.items,
                (intermediate.PrimitiveType, intermediate.OurTypeAnnotation),
            ), (
                "We expect only lists of primitive and our types. Lists of optionals "
                "and nested lists are not handled yet. Please contact the developers."
            )

            transformation_expression = _generate_transform_atomic_value(
                access_expression=Stripped("item"), type_anno=type_anno.items
            )

            var_name = typescript_naming.variable_name(Identifier(f"{prop.name}_array"))

            block = Stripped(
                f"""\
const {var_name} = new Array<JsonObject>();
for (const item of that.{prop_name}) {{
{I}{var_name}.push(
{II}{indent_but_first_line(transformation_expression, II)}
{I});
}}
jsonable[{key_literal}] = {var_name};"""
            )

        else:
            assert_never(type_anno)

        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            block = Stripped(
                f"""\
if (that.{prop_name} !== null) {{
{I}{indent_but_first_line(block, I)}
}}"""
            )

        blocks.append(block)

    if cls.serialization.with_model_type:
        model_type_literal = typescript_common.string_literal(
            naming.json_model_type(cls.name)
        )
        blocks.append(Stripped(f"""jsonable["modelType"] = {model_type_literal};"""))

    blocks.append(Stripped("return jsonable;"))

    method_name = typescript_naming.method_name(Identifier(f"transform_{cls.name}"))

    cls_name = typescript_naming.class_name(cls.name)

    writer = io.StringIO()

    writer.write(
        f"""\
/**
 * Serialize `that` to a JSON-able representation.
 *
 * @param that - instance to be serialization
 * @returns JSON-able representation
 */
{method_name}(
{I}that: AasTypes.{cls_name}
): JsonObject {{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_transformer(symbol_table: intermediate.SymbolTable) -> Stripped:
    methods = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        methods.append(_generate_transform(cls))

    writer = io.StringIO()
    writer.write(
        """\
/**
 * Transform the instance to its JSON-able representation.
 */
class Serializer extends AasTypes.AbstractTransformer<JsonObject> {
"""
    )

    for method in methods:
        writer.write("\n\n")
        writer.write(textwrap.indent(method, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


# endregion

# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the TypeScript code for the general de/serialization."""
    blocks = [
        typescript_description.documentation_comment(
            Stripped(
                """\
Provide de/serialization of AAS classes to/from JSON.

We can not use one-pass deserialization for JSON since the object
properties do not have fixed order, and hence we can not read
`modelType` property ahead of the remaining properties."""
            )
        ),
        typescript_common.WARNING,
        Stripped(
            """\
import * as AasCommon from "./common";
import * as AasTypes from "./types";
import * as AasStringification from "./stringification";"""
        ),
        Stripped(
            """\
export type JsonValue = string | number | boolean | JsonObject | JsonArray;

export type JsonArray = Iterable<JsonValue>;
export type JsonObject = { [prop: string]: JsonValue };"""
        ),
        Stripped(
            f"""\
/**
 * Represent a property on a path to the erroneous value.
 */
export class PropertySegment {{
{I}/**
{I} * Instance that contains the property
{I} */
{I}readonly instance: JsonObject;

{I}/**
{I} * Name of the property
{I} */
{I}readonly name: string;

{I}constructor(instance: JsonObject, name: string) {{
{II}this.instance = instance;
{II}this.name = name;
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Represent an index access on a path to the erroneous value.
 */
export class IndexSegment {{
{I}/**
{I} * Container that contains the item
{I} */
{I}readonly container: JsonArray;

{I}/**
{I} * Index of the item
{I} */
{I}readonly index: number;

{I}constructor(container: JsonArray, index: number) {{
{II}if (!Number.isInteger(index)) {{
{III}throw new Error(`Expected an integer for the index, but got: ${{index}}`);
{II}}}

{II}this.container = container;
{II}this.index = index;
{I}}}
}}"""
        ),
        Stripped(
            """\
export type Segment = PropertySegment | IndexSegment;"""
        ),
        Stripped(
            f"""\
/**
 * Represent the relative path to the erroneous value.
 */
export class Path {{
{I}private readonly _segments = new Array<Segment>();

{I}/**
{I} * Get the segments of the path.
{I} */
{I}segments(): Array<Segment> {{
{II}return this._segments;
{I}}}

{I}/**
{I} * Insert the `segment` in front of the {{@link segments}}.
{I} *
{I} * @param segment - segment to be prepended to {{@link segments}}
{I} */
{I}prepend(segment: Segment): void {{
{II}this._segments.unshift(segment);
{I}}}

{I}toString(): string {{
{II}if (this._segments.length === 0) {{
{III}return "";
{II}}}

{II}const parts = new Array<string>();

{II}let segment = this._segments[0];

{II}if (segment instanceof PropertySegment) {{
{III}parts.push(segment.name);
{II}}} else if (segment instanceof IndexSegment) {{
{III}parts.push(`[${{segment.index}}]`);
{II}}} else {{
{III}throw new Error(`Unexpected segment: ${{segment}}`);
{II}}}

{II}for (let i = 1; i < this._segments.length; i++) {{
{III}segment = this._segments[i];
{III}if (segment instanceof PropertySegment) {{
{IIII}parts.push(`.${{segment.name}}`);
{III}}} else if (segment instanceof IndexSegment) {{
{IIII}parts.push(`[${{segment.index}}]`);
{III}}} else {{
{IIII}throw new Error(`Unexpected segment: ${{segment}}`);
{III}}}
{II}}}

{II}return parts.join("");
{I}}}
}}"""
        ),
        Stripped("// region De-serialization"),
        Stripped(
            f"""\
/**
 * Signal that the JSON de-serialization could not be performed.
 */
export class DeserializationError {{
{I}/**
{I} * Human-readable explanation of the error
{I} */
{I}readonly message: string;

{I}/**
{I} * Relative path to the erroneous value
{I} */
{I}readonly path: Path;

{I}constructor(message: string, path: Path | null = null) {{
{II}this.message = message;
{II}this.path = path ?? new Path();
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Create an error as {{@link common.Either}}.
 *
 * @param message - human-readable explanation of the error
 * @returns An {{@link common.Either }} with the error set
 * @typeParam T - type of the value if there had been no error
 */
function newDeserializationError<T>(
{I}message: string
): AasCommon.Either<T, DeserializationError> {{
{I}return new AasCommon.Either<T, DeserializationError>(
{II}null,
{II}new DeserializationError(message)
{I});
}}"""
        ),
        _generate_bool_from_jsonable(),
        _generate_int_from_jsonable(),
        _generate_float_from_jsonable(),
        _generate_str_from_jsonable(),
        _generate_bytes_from_jsonable(),
    ]  # type: List[Stripped]

    errors = []  # type: List[Error]

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            blocks.append(_generate_enumeration_from_jsonable(enumeration=our_type))
        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            pass
        elif isinstance(our_type, intermediate.AbstractClass):
            blocks.append(
                _generate_dispatch_from_jsonable(interface=our_type.interface)
            )
        elif isinstance(our_type, intermediate.ConcreteClass):
            if len(our_type.concrete_descendants) > 0:
                assert our_type.interface is not None
                blocks.append(
                    _generate_dispatch_from_jsonable(interface=our_type.interface)
                )

            if our_type.is_implementation_specific:
                implementation_key = specific_implementations.ImplementationKey(
                    f"Jsonization/{our_type.name}_from_jsonable.ts"
                )

                implementation = spec_impls.get(implementation_key, None)
                if implementation is None:
                    errors.append(
                        Error(
                            our_type.parsed.node,
                            f"The jsonization snippet is missing "
                            f"for the implementation-specific "
                            f"class {our_type.name}: {implementation_key}",
                        )
                    )
                    continue
            else:
                # NOTE (mristin, 2022-11-25):
                # While TypeScript supports ``switch`` statement, it is not guaranteed
                # to run in sublinear time. Hence, we have to create a map with set
                # methods, see:
                # https://stackoverflow.com/questions/41109196/is-javascript-switch-statement-linear-or-constant-time
                blocks.append(_generate_setter(cls=our_type))

                blocks.append(_generate_concrete_class_from_jsonable(cls=our_type))
        else:
            assert_never(our_type)

    # NOTE (mristin, 2022-11-25):
    # We add all the dispatch mappings at the end as the functions might not have been
    # defined yet.
    for cls in symbol_table.classes:
        if isinstance(cls, intermediate.AbstractClass):
            blocks.append(_generate_dispatch_map_for_interface(interface=cls.interface))
        elif isinstance(cls, intermediate.ConcreteClass):
            if len(cls.concrete_descendants) > 0:
                assert (
                    cls.interface is not None
                ), "Expected an interface on a class with concrete descendants"

                blocks.append(
                    _generate_dispatch_map_for_interface(interface=cls.interface)
                )

            if not cls.is_implementation_specific:
                blocks.append(_generate_setter_map(cls=cls))

        else:
            assert_never(cls)

    blocks.append(Stripped("// endregion"))

    blocks.append(Stripped("// region Serialization"))

    blocks.append(_generate_transformer(symbol_table=symbol_table))

    blocks.append(Stripped("const SERIALIZER = new Serializer();"))

    # pylint: disable=line-too-long
    blocks.append(
        Stripped(
            f"""\
/**
 * Convert `that` to a JSON-able structure.
 *
 * @param that - AAS data to be recursively converted to a JSON-able structure
 * @returns
 * JSON-able structure which can be further processed with, say,
 * {{@link https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/JSON/stringify|JSON.stringify}})
 */
export function toJsonable(that: AasTypes.Class): JsonObject {{
{I}return SERIALIZER.transform(that);
}}"""
        )
    )
    # pylint: enable=line-too-long

    blocks.append(Stripped("// endregion"))

    if len(errors) > 0:
        return None, errors

    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None



"""Generate C# code for copying in memory based on the intermediate representation."""

import io
import textwrap
from typing import Tuple, Optional, List

from icontract import ensure

from aas_core_codegen import intermediate, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    Identifier,
    indent_but_first_line,
    assert_never,
)
from aas_core_codegen.csharp import (
    common as csharp_common,
    naming as csharp_naming,
)
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
)


def _generate_shallow_copy_transform_method(
    cls: intermediate.ConcreteClass,
) -> Stripped:
    """Generate the method in the transformer to make a shallow copy of ``cls``."""
    property_names = [prop.name for prop in cls.properties]
    constructor_argument_names = [arg.name for arg in cls.constructor.arguments]

    # fmt: off
    assert (
            set(prop.name for prop in cls.properties)
            == set(arg.name for arg in cls.constructor.arguments)
    ), (
        f"Expected the properties to coincide with constructor arguments, "
        f"but they do not for {cls.name!r}:"
        f"{property_names=}, {constructor_argument_names=}"
    )
    # fmt: on

    cls_name = csharp_naming.class_name(cls.name)

    if len(cls.constructor.arguments) == 0:
        return_statement = Stripped(f"return new Aas.{cls_name}();")
    else:
        constructor_arg_exprs = []  # type: List[str]
        for arg in cls.constructor.arguments:
            prop_name = csharp_naming.property_name(arg.name)
            constructor_arg_exprs.append(f"that.{prop_name}")

        # NOTE (mristin, 2022-11-03):
        # This is poor man's heuristic for line breaking, but it works fairly well
        # in practice.
        args_joined = ", ".join(constructor_arg_exprs)
        return_statement = Stripped(f"return new Aas.{cls_name}({args_joined});")

        if len(return_statement) > 70:
            args_joined = ",\n".join(constructor_arg_exprs)
            return_statement = Stripped(
                f"""\
return new Aas.{cls_name}(
{I}{indent_but_first_line(args_joined, I)});"""
            )

    interface_name = csharp_naming.interface_name(cls.name)
    transform_name = csharp_naming.method_name(Identifier(f"transform_{cls.name}"))

    return Stripped(
        f"""\
public override Aas.IClass {transform_name}(
{I}Aas.{interface_name} that
)
{{
{I}{indent_but_first_line(return_statement, I)}
}}"""
    )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_shallow_copier(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the transformer which makes shallow copies."""
    errors = []  # type: List[Error]

    blocks = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        if cls.is_implementation_specific:
            implementation_key = specific_implementations.ImplementationKey(
                f"Copying/ShallowCopier/transform_{cls.name}.cs"
            )

            implementation = spec_impls.get(implementation_key, None)
            if implementation is None:
                errors.append(
                    Error(
                        cls.parsed.node,
                        f"The snippet for making shallow copies is missing "
                        f"for the implementation-specific "
                        f"class {cls.name}: {implementation_key}",
                    )
                )
                continue

            blocks.append(spec_impls[implementation_key])
        else:
            blocks.append(_generate_shallow_copy_transform_method(cls=cls))

    writer = io.StringIO()
    writer.write(
        """\
/// <summary>Dispatch the making of shallow copies.</summary>
internal class ShallowCopier : Visitation.AbstractTransformer<Aas.IClass>
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // internal class ShallowCopier")

    return Stripped(writer.getvalue()), None


def _generate_deep_copy_transform_method(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the method in the transformer to make a deep copy of ``cls``."""
    property_names = [prop.name for prop in cls.properties]
    constructor_argument_names = [arg.name for arg in cls.constructor.arguments]

    # fmt: off
    assert (
            set(prop.name for prop in cls.properties)
            == set(arg.name for arg in cls.constructor.arguments)
    ), (
        f"Expected the properties to coincide with constructor arguments, "
        f"but they do not for {cls.name!r}:"
        f"{property_names=}, {constructor_argument_names=}"
    )
    # fmt: on

    cls_name = csharp_naming.class_name(cls.name)

    body_blocks = []  # type: List[Stripped]

    if len(cls.constructor.arguments) == 0:
        body_blocks.append(Stripped(f"return new Aas.{cls_name}();"))
    else:
        # NOTE (mristin, 2022-11-04):
        # We could use LINQ to make deep copies of lists in expressions which are
        # directly passed to the constructor. This indeed makes sense if we wrote
        # the code manually, and would also be a more elegant solution. However,
        # LINQ comes with a certain overhead (for example, the memory for the new list
        # can not be pre-reserved ahead of time). That is why we make these copies in
        # separate variables and pass on the variables to the constructor.
        for arg in cls.constructor.arguments:
            prop_name = csharp_naming.property_name(arg.name)
            optional = isinstance(
                arg.type_annotation, intermediate.OptionalTypeAnnotation
            )

            type_anno = intermediate.beneath_optional(arg.type_annotation)

            if not isinstance(type_anno, intermediate.ListTypeAnnotation):
                continue

            assert isinstance(
                type_anno.items, intermediate.OurTypeAnnotation
            ) and isinstance(type_anno.items.our_type, intermediate.Class), (
                "(mristin, 2022-11-03) We handle only lists of classes in the deep "
                "copies at the moment. The meta-model does not contain "
                "any other lists, so we wanted to keep the code as simple as "
                "possible, and avoid unrolling. Please contact the developers "
                "if you need this feature."
            )

            # NOTE (mristin, 2022-11-04):
            # We need to prefix to avoid any possible naming conflicts.
            variable_name = csharp_naming.variable_name(Identifier(f"the_{arg.name}"))

            variable_type = csharp_common.generate_type(type_anno)

            if not optional:
                body_blocks.append(
                    Stripped(
                        f"""\
var {variable_name} = new {variable_type}(
{I}that.{prop_name}.Count);
foreach (var item in that.{prop_name})
{{
{I}{variable_name}.Add(Deep(item));
}}"""
                    )
                )
            else:
                body_blocks.append(
                    Stripped(
                        f"""\
{variable_type}? {variable_name} = null;
if (that.{prop_name} != null)
{{
{I}{variable_name} = new {variable_type}(
{II}that.{prop_name}.Count);
{I}foreach (var item in that.{prop_name})
{I}{{
{II}{variable_name}.Add(Deep(item));
{I}}}
}}"""
                    )
                )

        constructor_arg_exprs = []  # type: List[str]
        for arg in cls.constructor.arguments:
            prop_name = csharp_naming.property_name(arg.name)

            type_anno = intermediate.beneath_optional(arg.type_annotation)

            optional = isinstance(
                arg.type_annotation, intermediate.OptionalTypeAnnotation
            )

            if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
                constructor_arg_exprs.append(f"that.{prop_name}")

            elif isinstance(type_anno, intermediate.OurTypeAnnotation):
                if isinstance(type_anno.our_type, intermediate.Enumeration):
                    constructor_arg_exprs.append(f"that.{prop_name}")

                elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
                    constructor_arg_exprs.append(f"that.{prop_name}")

                elif isinstance(
                    type_anno.our_type,
                    (intermediate.AbstractClass, intermediate.ConcreteClass),
                ):
                    if optional:
                        constructor_arg_exprs.append(
                            f"""\
(that.{prop_name} != null)
{I}? Deep(that.{prop_name})
{I}: null"""
                        )
                    else:
                        constructor_arg_exprs.append(f"Deep(that.{prop_name})")
                else:
                    # noinspection PyTypeChecker
                    assert_never(type_anno.our_type)

            elif isinstance(type_anno, intermediate.ListTypeAnnotation):
                # NOTE (mristin, 2022-11-04):
                # See how this variable is computed above in the generated code.
                constructor_arg_exprs.append(
                    csharp_naming.variable_name(Identifier(f"the_{arg.name}"))
                )
            else:
                # noinspection PyTypeChecker
                assert_never(type_anno)

        return_statement_writer = io.StringIO()

        return_statement_writer.write(f"return new Aas.{cls_name}(\n")

        for i, arg_expr in enumerate(constructor_arg_exprs):
            return_statement_writer.write(textwrap.indent(arg_expr, I))

            if i < len(constructor_arg_exprs) - 1:
                return_statement_writer.write(",\n")
            else:
                return_statement_writer.write("\n")

        return_statement_writer.write(");")

        body_blocks.append(Stripped(return_statement_writer.getvalue()))

    body_writer = io.StringIO()
    for i, body_block in enumerate(body_blocks):
        if i > 0:
            body_writer.write("\n\n")

        body_writer.write(body_block)

    interface_name = csharp_naming.interface_name(cls.name)
    transform_name = csharp_naming.method_name(Identifier(f"transform_{cls.name}"))

    return Stripped(
        f"""\
public override Aas.IClass {transform_name}(
{I}Aas.{interface_name} that
)
{{
{I}{indent_but_first_line(body_writer.getvalue(), I)}
}}"""
    )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_deep_copier(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the transformer which makes deep copies."""
    errors = []  # type: List[Error]

    blocks = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        if cls.is_implementation_specific:
            implementation_key = specific_implementations.ImplementationKey(
                f"Copying/DeepCopier/transform_{cls.name}.cs"
            )

            implementation = spec_impls.get(implementation_key, None)
            if implementation is None:
                errors.append(
                    Error(
                        cls.parsed.node,
                        f"The snippet for making deep copies is missing "
                        f"for the implementation-specific "
                        f"class {cls.name}: {implementation_key}",
                    )
                )
                continue

            blocks.append(spec_impls[implementation_key])
        else:
            blocks.append(_generate_deep_copy_transform_method(cls=cls))

    writer = io.StringIO()
    writer.write(
        """\
/// <summary>Dispatch the making of deep copies.</summary>
internal class DeepCopier : Visitation.AbstractTransformer<Aas.IClass>
{"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n}  // internal class DeepCopier")

    return Stripped(writer.getvalue()), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    namespace: csharp_common.NamespaceIdentifier,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the C# code for copying instances in memory.

    The ``namespace`` defines the AAS C# namespace.
    """
    errors = []  # type: List[Error]

    using_directives = []  # type: List[Stripped]
    using_directives.extend(
        csharp_common.generate_using_aas_directive_if_necessary(namespace)
    )

    using_directives.append(
        Stripped(
            """\
using System.Collections.Generic;  // can't alias"""
        )
    )

    # NOTE (mristin, 2022-11-03):
    # We wrap the shallow and deep copying in generic methods to allow for easier
    # enforcement of runtime type safety for the client. Otherwise, if we directly
    # provided the transformer, the client would always need to make the casts, which
    # is cumbersome.

    copy_blocks = [
        Stripped(
            f"""\
private static readonly ShallowCopier ShallowCopierInstance = (
{I}new ShallowCopier());"""
        ),
        Stripped(
            f"""\
private static readonly DeepCopier DeepCopierInstance = (
{I}new DeepCopier());"""
        ),
        Stripped(
            f"""\
/// <summary>
/// Make a shallow copy of <paramref name="that" />.
/// </summary>
/// <remarks>
/// All the properties are copied by reference. This includes also the lists.
/// Hence, a list property is copied by reference, and not, as sometimes might be
/// expected, as a new list of underlying references.
/// </remarks>.
/// <param name="that">to be copied in a shallow manner</param>
/// <typeparam name="T">type to cast the result to</typeparam>
public static T Shallow<T>(T that) where T : Aas.IClass
{{
{I}return (T)ShallowCopierInstance.Transform(that);
}}"""
        ),
        Stripped(
            f"""\
/// <summary>
/// Make a recursively a deep copy of <paramref name="that" />.
/// </summary>
/// <param name="that">to be deeply copied in a recursive manner</param>
/// <typeparam name="T">type to cast the result to</typeparam>
public static T Deep<T>(T that) where T : Aas.IClass
{{
{I}return (T)DeepCopierInstance.Transform(that);
}}"""
        ),
    ]  # type: List[Stripped]

    shallow_copier_block, shallow_errors = _generate_shallow_copier(
        symbol_table=symbol_table, spec_impls=spec_impls
    )
    if shallow_errors is not None:
        errors.extend(shallow_errors)
    else:
        assert shallow_copier_block is not None
        copy_blocks.append(shallow_copier_block)

    deep_copier_block, deep_errors = _generate_deep_copier(
        symbol_table=symbol_table, spec_impls=spec_impls
    )
    if deep_errors is not None:
        errors.extend(deep_errors)
    else:
        assert deep_copier_block is not None
        copy_blocks.append(deep_copier_block)

    if len(errors) > 0:
        return None, errors

    copying_writer = io.StringIO()
    copying_writer.write(
        """\
/// <summary>
/// Allow for making shallow and deep copies of AAS model instances.
/// </summary>
public static class Copying
{
"""
    )

    for i, copy_block in enumerate(copy_blocks):
        if i > 0:
            copying_writer.write("\n\n")

        copying_writer.write(textwrap.indent(copy_block, I))

    copying_writer.write("\n}  // public static class Copying")

    blocks = [
        csharp_common.WARNING,
        Stripped("\n".join(using_directives)),
        Stripped(
            f"""\
namespace {namespace}
{{
{I}{indent_but_first_line(copying_writer.getvalue(), I)}
}}  // namespace {namespace}"""
        ),
        csharp_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        assert not block.startswith("\n")
        assert not block.endswith("\n")
        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None
